#ifndef _BELIEF_PROPAGATION_MS_H
#define _BELIEF_PROPAGATION_MS_H

/// vertex assignment
///\param n number of levels
///\param N number of vertex
///\param **dst is the array of destianations
#ifdef FLOAT
void assign_vertex(int n, int N, double **dst);
#else
void assign_vertex(int n, int N, int **dst);
#endif // FLOAT
/// compute remap sequence
///\param *remap is the remapping sequence
///\param N is the remapping sequence length
void get_remap(int *remap, int N);
void reverse_int(int *temp, int dim);

/// Backward message computation
///\param *left is the left message array
///\param *right is the right message array
///\param **dst is the destination array
///\param n=log2(N)
///\param N is the coded size
///\param min_val_ex is the minimum value for the extrinsic information
///\param max_val_ex is the maximum value for the extrinsic information
///\param sfact is the scaling factor to implement normalized min sum
//void left_message(int **msg, int **dst, int n, int N, int min_val_ex, int max_val_ex, float sfact);
void left_message_ms(int *left, const int *right, int **dst, int n, int N, int min_val_ex, int max_val_ex, float sfact);

/// Forward message computation
///\param *left is the left message array
///\param *right is the right message array
///\param **dst is the destination array
///\param n=log2(N)
///\param N is the coded size
///\param min_val_ex is the minimum value for the extrinsic information
///\param max_val_ex is the maximum value for the extrinsic information
///\param sfact is the scaling factor to implement normalized min sum
void right_message_ms(const int *left, int *right, int **dst, int n, int N, int min_val_ex, int max_val_ex, float sfact);

/// matrices initialization
///\param *left is the left message array
///\param *right is the right message array
///\param *llr is log likelihood ratios from the channel
///\param *index_b is the list of bad channels
///\param *index_g is the list of good channels
///\param nN is log(N)*N
///\param N is the coded size
///\param K is the uncoded size
///\param max_val_ex is the maximum value for the extrinsic information
#ifdef FLOAT
void init_matrix_ms(double *left, double *right, double *llr,int *index_b, int *index_g, int nN, int N, int K,int max_val_ex);
#else
void init_matrix_ms(int *left, int *right, int *llr,int *index_b, int *index_g, int nN, int N, int K,int max_val_ex);
#endif // FLOAT

#endif
