#include<math.h>
#include <float.h>
#include"myservs.h"
#include"belief_propagation_sorted.h"
#include"belief_propagation_reduced.h"



/// compute the minimum as in LDCP min-sum algorithm:
/// sign is the product of signs
/// magnitue is the minimum magnitude
///\param a is the first input
///\param b is the recond input
///\param s is the scaling factor
///\return the min-sum result
static int min_sum(int a, int b, float s)
{
  register int sa, sb;
  register int ma, mb;
  register int d;
  register int sm;
  register int m;

  /// get the signs
  sa = MY_INT_SIGN(a);
  sb = MY_INT_SIGN(b);
  /// compute result sign
  sm = sa ^ sb;
  /// get magnitudes
  ma = sa ^ a;
  ma += sa & 0x1;
  mb = sb ^ b;
  mb += sb & 0x1;
  /// find the min
  d = ma - mb;
  m = (d >= 0) ? mb : ma;
  /// apply the sign
  m ^= sm;
  m += sm & 0x1;
  /// apply scaling
  m = (int)floor(s * m + 0.5);

  return m;
}


static int clip(int a,int min_val_ex, int max_val_ex)
{
    if(a>max_val_ex)
        return max_val_ex;
    if(a<min_val_ex)
        return min_val_ex;

    return a;
}



/// Forward and backward message computation
///\param *left is the left message array
///\param *right is the right message array
///\param **dst is the destination array
///\param n=log2(N)
///\param N is the coded size
///\param min_val_ex is the minimum value for the extrinsic information
///\param max_val_ex is the maximum value for the extrinsic information
///\param sfact is the scaling factor to implement normalized min sum
//void right_message(int **msg, int **dst, int n, int N, int min_val_ex, int max_val_ex, float sfact)

void left_right_message_ms_sorted(int *left, int *right, int *temp, int n, int N, int min_val_ex, int max_val_ex, float sfact)
#ifndef FLOAT
{
  int i,j,k;
  int a;
  int b;
  int r1;
  int r2;

  for (i=0; i<n; i++)
  {
    k = i*N;
    backsorter(left,i+1,N,temp);
    for (j=0; j<N; j+=2)
    {
     /// CHECK NODE
      /// get messages
      a = left[k+N];
      b = left[k+N+1] + right[k+1];
      b=clip(b, min_val_ex, max_val_ex);
      /// compute min sum
      r1 = min_sum(a, b, sfact);
      r1=clip(r1, min_val_ex, max_val_ex);

      /// VARIABLE NODE
      a = left[k+N];
      b = right[k];
      /// compute min sum
      r2 = min_sum(a, b, sfact);
      r2 += left[k+1+N]; /// add the left value of second next node destination to Min Sum
      r2=clip(r2, min_val_ex, max_val_ex);

      /// save output values
      left[k] = r1;   /// L(i,j)
      left[k+1] = r2; /// L(i,j+Ni)

      //printf("%d %d %d %d %d %d\n", k, k+1, dst[k][0], dst[k][1], dst[k+1][0], dst[k+1][1]);

      /// CHECK NODE
      /// get messages
      a = right[k];
      b = left[k+1+N] + right[k+1];
      b=clip(b, min_val_ex, max_val_ex);
      /// min sum
      r1 = min_sum(a, b, sfact);
      r1=clip(r1, min_val_ex, max_val_ex);

      /// VARIABLE NODE
      /// get messages
      a = right[k];
      b = left[k+N];
      r2 = min_sum(a, b, sfact);
      r2 += right[k+1];//add the right value of next node to min sum
      r2=clip(r2, min_val_ex, max_val_ex);

      right[k+N] = r1; /// R(i+1,j)
      right[k+N+1] = r2; /// R(i+1,j+N)


      k += 2;
    }
    sorter(right,i+1,N,temp);
    ///if(i==n-1)
    sorter(left,i+1,N,temp);
  }


//  for (i=N; i<n+N; i++)
//  right[i]=0;
}
#else
{
    printf("Error float mode non compatible");
    exit(1);
}
#endif // FLOAT


/// Forward and backward message computation
///\param *left is the left message array
///\param *right is the right message array
///\param **dst is the destination array
///\param n=log2(N)
///\param N is the coded size
///\param min_val_ex is the minimum value for the extrinsic information
///\param max_val_ex is the maximum value for the extrinsic information
///\param sfact is the scaling factor to implement normalized min sum
//void right_message(int **msg, int **dst, int n, int N, int min_val_ex, int max_val_ex, float sfact)

void left_right_message_ms_sorted2(int *left, int *right, int *temp, int n, int N, int min_val_ex, int max_val_ex, float sfact)
#ifndef FLOAT
{
  int i,j,k;
  int a;
  int b;
  int r1;
  int r2;

  for (i=0; i<n; i++)
  {
    k = i*N;
    sorter(right,i,N,temp);
    for (j=0; j<N; j+=2)
    {
     /// CHECK NODE
      /// get messages
      a = left[k+N];
      b = left[k+N+1] + right[k+1];
      b=clip(b, min_val_ex, max_val_ex);
      /// compute min sum
      r1 = min_sum(a, b, sfact);
      r1=clip(r1, min_val_ex, max_val_ex);

      /// VARIABLE NODE
      a = left[k+N];
      b = right[k];
      /// compute min sum
      r2 = min_sum(a, b, sfact);
      r2 += left[k+1+N]; /// add the left value of second next node destination to Min Sum
      r2=clip(r2, min_val_ex, max_val_ex);

      /// save output values
      left[k] = r1;   /// L(i,j)
      left[k+1] = r2; /// L(i,j+Ni)

      //printf("%d %d %d %d %d %d\n", k, k+1, dst[k][0], dst[k][1], dst[k+1][0], dst[k+1][1]);

      /// CHECK NODE
      /// get messages
      a = right[k];
      b = left[k+1+N] + right[k+1];
      b=clip(b, min_val_ex, max_val_ex);
      /// min sum
      r1 = min_sum(a, b, sfact);
      r1=clip(r1, min_val_ex, max_val_ex);

      /// VARIABLE NODE
      /// get messages
      a = right[k];
      b = left[k+N];
      r2 = min_sum(a, b, sfact);
      r2 += right[k+1];//add the right value of next node to min sum
      r2=clip(r2, min_val_ex, max_val_ex);

      right[k+N] = r1; /// R(i+1,j)
      right[k+N+1] = r2; /// R(i+1,j+N)


      k += 2;
    }
    backsorter(right,i,N,temp);
    ///if(i==n-1)
    backsorter(left,i,N,temp);
  }


//  for (i=N; i<n+N; i++)
//  right[i]=0;
}
#else
{
    printf("Error float mode non compatible");
    exit(1);
}
#endif // FLOAT


void left_right_message_ms_sorted3(int *left, int *right, int *temp, int n, int N, int min_val_ex, int max_val_ex, float sfact)
#ifndef FLOAT
{
  int i,j,k;
  int a;
  int b;
  int r1;
  int r2;

  for (i=n-1; i>=0; i--)
  {
    k = i*N;
    sorter(right,i,N,temp);
    for (j=0; j<N; j+=2)
    {
     /// CHECK NODE
      /// get messages
      a = left[k+N];
      b = left[k+N+1] + right[k+1];
      b=clip(b, min_val_ex, max_val_ex);
      /// compute min sum
      r1 = min_sum(a, b, sfact);
      r1=clip(r1, min_val_ex, max_val_ex);

      /// VARIABLE NODE
      a = left[k+N];
      b = right[k];
      /// compute min sum
      r2 = min_sum(a, b, sfact);
      r2 += left[k+1+N]; /// add the left value of second next node destination to Min Sum
      r2=clip(r2, min_val_ex, max_val_ex);

      /// save output values
      left[k] = r1;   /// L(i,j)
      left[k+1] = r2; /// L(i,j+Ni)

      //printf("%d %d %d %d %d %d\n", k, k+1, dst[k][0], dst[k][1], dst[k+1][0], dst[k+1][1]);

      /// CHECK NODE
      /// get messages
      a = right[k];
      b = left[k+1+N] + right[k+1];
      b=clip(b, min_val_ex, max_val_ex);
      /// min sum
      r1 = min_sum(a, b, sfact);
      r1=clip(r1, min_val_ex, max_val_ex);

      /// VARIABLE NODE
      /// get messages
      a = right[k];
      b = left[k+N];
      r2 = min_sum(a, b, sfact);
      r2 += right[k+1];//add the right value of next node to min sum
      r2=clip(r2, min_val_ex, max_val_ex);

      right[k+N] = r1; /// R(i+1,j)
      right[k+N+1] = r2; /// R(i+1,j+N)


      k += 2;
    }
    backsorter(right,i,N,temp);
    ///if(i==n-1)
    backsorter(left,i,N,temp);
  }


//  for (i=N; i<n+N; i++)
//  right[i]=0;
}
#else
{
    printf("Error float mode non compatible");
    exit(1);
}
#endif // FLOAT
