/// Polar codes model 1.1
/// decoder is based on Belief propagation
/// Authors: A. Di Paolo/M.B. Scaraffia/G. Masera/M. Martina
/// - norm min differatiation CN/VN
/// - check saturation as if in LDPC model and apply E. Boutillon solution
//-funroll-loops -ffast-math -Ofast -flto -m64 -march=native
#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<time.h>
#include <float.h>
#include"myretcodes.h"
#include"myservs.h"
#include"params.h"
#include"channels.h"
#include"belief_propagation_full.h"
#include"belief_propagation_full_v2.h"
#include"belief_propagation_ms.h"
#include"belief_propagation_ms_v2.h"
#include"belief_propagation_reduced.h"
#include"belief_propagation_sorted.h"
#include"belief_propagation_generic.h"
#include"random_sel.h"
#include"encoder.h"
#include"results.h"
#include"modulate2.h"

#include"crc32.c"
#include"crc16.c"
#include<string.h>

#define _CRT_SECURE_NO_WARNINGS
#define BPSK_G 2.0

int main (int argc, char **argv)
{
  int error_stop=0;
  int opcount = 0;
  int op_per_iter = 0;
  FILE *llr;
  int *iteration_reached;
  //int zz = 0;
  //FILE *testfp;
  FILE *fp_param; /// parameter file
  FILE *fp_res; /// results file
  FILE *fp_res_det; /// results file

  FILE *fp_res_matrix; /// results file
  FILE *fp_res_in; /// input file
  FILE *fp_res_out; /// output file

  FILE *fp_setup; /// file di setup dellal memoria right

  params_t sim_par;

  char *oname; /// pointer to the output file name
  char tmpbuf[LARGE_INT_VAL]; /// temporary buffer

  int K; /// uncoded size
  int N; /// code size
  int n; /// log2(N)
  double rate; /// code rate K/N

  int curr_frame; /// current frame index
  int max_frame; /// maximum number of simulated frames
  int max_err; /// maximum number of wrong frames counted

  unsigned int lfsr_seed; /// seed for the LFSR
  unsigned int unif_seed; /// seed for the uniform distribution

  int curr_point; /// current SNR point
  int min_point; /// minimum SNR point
  int max_point; /// maximum SNR point

  double ebn0_min; /// minimum SNR value
  double ebn0_step; /// SNR step

  int curr_iter; /// current iteration
  int min_iter; /// minimum number of iterations
  int max_iter; /// maximum number of iterations
  int *avg_iter; /// average number of iterations for statistic purpose

  int *tot_frames; /// total number of frames counter
  int **bit_err; /// error bit counter
  int **frame_err; /// error frame counter

  int max_point_comp; /// maximum SNR point computed


  int min_val_int; /// minimum value for intrinsic information
  int max_val_int; /// maximum value for intrinsic information
  int frac_bits; /// number of fractional bit for LLR generation
  int min_val_ext; /// minimum value for extrinsic information
  int max_val_ext; /// maximum value for extrinsic information
  int r_init;
  float sfact_left; /// scaling factor for min-sum algorithm left messages
  float sfact_right; /// scaling factor for min-sum algorithm right messages

  clock_t ck_1, ck_2, ck_delta, ck_max; /// clock variable for time display
  int time_check; /// check time amount for display

  int f_num; /// first frame number

  double bhatta_eps; /// Bhattacharyya epsilon
  channel_t *w; /// channels

  /// graph matrices
#ifdef FLOAT
  double *matrix_messages_left;
  double *matrix_messages_right;
  double **matrix_destinations;

  double *temp_red;

#else
  int *matrix_messages_left;
  int *matrix_messages_right;
  int **matrix_destinations;
  double *matrix_messages_left_d;
  double *matrix_messages_right_d;

  int *temp_red;
#endif // FLOAT

  /// encoder side array
  int *uncodedK;
  int *uncodedN;
  int *coded;
  int *recoded;
  int *channelX;

  int *FBER, FBER_temp, FBER_count, FBER_M, *dec_FBER, M, FBER_stat=0;
  float FBER_stat_per;
  int tot_stop_frame;

  int crc_case, crc_count=0, crc_stat=0;
  float crc_stat_per;
  unsigned char *crc_input_coded, *crc_input_uncoded;
  uint32_t *crc32_coded, *crc32_uncoded;
  uint16_t *crc16_coded, *crc16_uncoded;

  int *index_g;
  int *index_b;
  int **G, G_stat=0;
  float G_stat_per;
  int **Gt;
  ///int *remap; /// permutation sequence for correct BP feeding

  double *noise; /// noise array
  double level; /// noise std deviation
  double modulated; /// modulated value
  double received; /// received value
#ifdef FLOAT
  double *llr_in; /// intrinsic information array
#else
  int *llr_in; /// intrinsic information array
#endif // FLOAT
  double *llr_ch; /// intrinsic information array
  int *dec; /// hard decoding bits
  int *decoded; /// hard decoding bits
  int *decodedK; /// hard decoding bits

  int nN; /// n*N
  int errors; /// number of errors, that is BERmeter return value to stop iterations

  int i; /// general purpose index
    int dataindex;

  generic_f f_generico;/// per decoding con funzione generica
  generic_f f_gen_app;/// per decoding con funzione generica con partenza appoggio.

///////////////////////////////////////////////////////////////////////////////////////////
///VARIABILI DI CONFIGURAZIONE DELLA SIMULAZIONE
  int decoding_type; /// type of decoding to use
  int frame_by_frame=0;///save every frame information-- per autocheck-- ricorda il breakpoint!!!
  int hw_check_en=0;///uncoded data contains ones
  int save_details_en=0;///save every frame information-- per valutazioni con matlab
  int hw_autotest_file=0;///tipo di output compresso-- per autotest
  int save_hw_initial=0; ///per salvare i dati in memoria iniziali di generic linear (funziona solo per decoding_type=18)
  int BEC_AWGN_choice=1;///se 0 BEC, se 1 AWGN per i frozen bit

  int QAM_number=16;/// se 0 BPSK, altrimenti 16-64-256 QAM
  int bicm=1;

  int stopping_rule;///select the stopping rule to use -> 0 none, 1 break quando no errori, 2 break quando trovati frozen bit, 3 Gmatrix, 5 FBER
///////////////////////////////////////////////////////////////////////////////////////////
  random_seed rndseedlfsr;
  random_seed rndseedunif;
  rndseedlfsr.seed=&lfsr_seed;
  rndseedunif.seed=&unif_seed;
  int swiching_state=0;
  int continue_decoding=1;///per stopping_rule =2
  int exit_ite_loop=0;
  //aggiunto
   float acc_time=0;
   float velocity;
//
  int* interleave_bcim;

  #ifdef FLOAT
  printf("WARNING FLOAT MODE ENABLED!\n");
  #endif // FLOAT
//  testfp = fopen("messages.txt","w");
  if (argc != 2)
  {
    printf("Use: %s <param_file>\n", argv[0]);
    //exit(KO_USAGE);

  }

  /// check architecture to be sure about int representation ...
  if (sizeof(int) != MY_INT_SIZE)
  {
    printf("This architecture has sizeof(int)=%u instead of %u\n", sizeof(int), MY_INT_SIZE);
    printf("Please change MY_INT_SIZE in myservs.h to support your architecture\n");
    exit(KO_INT_SIZE);
  }

  /// check architecture for clock_t type
  if (sizeof(clock_t) != MY_CLOCK_T_SIZE)
  {
    printf("This architecture has sizeof(clock_t)=%d instead of %d\n", sizeof(clock_t), MY_CLOCK_T_SIZE);
    printf("Please change MY_CLOCK_T_SIZE and MY_CLOCK_T_MAX in myservs.h to support your architecture\n");
    exit(KO_CLOCK_T_SIZE);
  }

  /// get parameters from file
  fp_param = myfopen(argv[1], "r", "main");
  get_params(fp_param, &sim_par);
  fclose(fp_param);

  /// copy parameters to automatic variables
  K = sim_par.K;
  N = sim_par.N;
  max_frame = sim_par.max_frame;
  max_err = sim_par.max_err;
  ebn0_min =  sim_par.ebn0_min;
  ebn0_step = sim_par.ebn0_step;
  max_point = sim_par.max_point;
  min_point = sim_par.min_point;
  max_point_comp = sim_par.max_point_comp;
  min_iter = sim_par.min_iter;
  max_iter = sim_par.max_iter;
  oname = sim_par.oname;
  time_check = sim_par.time_check;
  lfsr_seed = sim_par.lfsr_seed;
  unif_seed = sim_par.unif_seed;
  f_num = sim_par.f_num;
  tot_frames = sim_par.tot_frames;
  bit_err = sim_par.bit_err;
  frame_err = sim_par.frame_err;
  avg_iter = sim_par.avg_iter;
  bhatta_eps = sim_par.bhatta_eps;
  min_val_int = sim_par.min_val_int;
  max_val_int = sim_par.max_val_int;
  frac_bits = sim_par.frac_bits;
  min_val_ext = sim_par.min_val_ext;
  max_val_ext = sim_par.max_val_ext;
  r_init = sim_par.R_init;
  sfact_left = sim_par.sfact_left;
  sfact_right = sim_par.sfact_right;
  decoding_type = sim_par.decoding_type_number;
  QAM_number = sim_par.QAM_modulation_number;
  bicm = sim_par.BICM_enable;
  stopping_rule=sim_par.stopping;
  rndseedlfsr.sel = &sim_par.lfsr_sel;
  rndseedunif.sel = &sim_par.unif_sel;

  setseed(&rndseedunif);
  setseed(&rndseedlfsr);

  /// set local variables
  rate = (double)K/N;
  n = (int)ceil(log2(N));
  nN = n*N;

  /// show parameters
  show_params(sim_par, rate);

  /// allocate local arrays
   w = (channel_t*) myalloc(N, sizeof(channel_t), 1, "main", "w");
  if(decoding_type!=5)
  {
  #ifndef FLOAT
  matrix_destinations = (int**) myalloc((n+1)*N, sizeof(int*), 0, "main", "matrix_destinations");
  for (i=0; i<(n+1)*N; i++)
  matrix_destinations[i] = (int*) myalloc(2, sizeof(int), 1, "main", "matrix_destinations[i]");
  }
  #else
  matrix_destinations = (double**) myalloc((n+1)*N, sizeof(double*), 0, "main", "matrix_destinations");
  for (i=0; i<(n+1)*N; i++)
  matrix_destinations[i] = (double*) myalloc(2, sizeof(double), 1, "main", "matrix_destinations[i]");
  }
  #endif
  switch(decoding_type)
        {
            #ifndef FLOAT
            case 0 :
            case 2 :
            case 3 :matrix_messages_left_d = (double*) myalloc((n+1)*N, sizeof(double), 1, "main", "matrix_messages_left_double");
                    matrix_messages_right_d = (double*) myalloc((n+1)*N, sizeof(double), 1, "main", "matrix_messages_right_double");
                    break;
            case 1 :
            case 4 :matrix_messages_left = (int*) myalloc((n+1)*N, sizeof(int), 1, "main", "matrix_messages_left");
                    matrix_messages_right = (int*) myalloc((n+1)*N, sizeof(int), 1, "main", "matrix_messages_right");
                    break;
            case 5:
            case 8: matrix_messages_left = (int*) myalloc((n+1)*N, sizeof(int), 1, "main", "matrix_messages_left");
                    matrix_messages_right = (int*) myalloc(2*N, sizeof(int), 1, "main", "matrix_messages_right");
                    temp_red=(int*) myalloc(N, sizeof(int), 1, "bp_reduced", "sorter");
                    break;
            case 6 :
            case 9 :
            case 10:
            case 11:
            case 12:
            case 13:
            case 7 :matrix_messages_left = (int*) myalloc((n+1)*N, sizeof(int), 1, "main", "matrix_messages_left");
                    matrix_messages_right = (int*) myalloc((n+1)*N, sizeof(int), 1, "main", "matrix_messages_right");
                    temp_red=(int*) myalloc(N, sizeof(int), 1, "bp_sorted", "sorter");
                    break;
            #endif // FLOAT
            case 14:
            case 15:
            case 16:
            case 17:
            case 18:
            case 19:
            case 20:
                    #ifdef FLOAT
                    matrix_messages_left = (double*) myalloc((n+1)*N, sizeof(double), 1, "main", "matrix_messages_left");
                    matrix_messages_right = (double*) myalloc((n+1)*N, sizeof(double), 1, "main", "matrix_messages_right");
                    temp_red=(double*) myalloc(N, sizeof(double), 1, "bp_sorted", "sorter");
                    f_generico.left_temp= (double*) myalloc((n+1)*N, sizeof(double), 1, "main", "matrix_messages_left_temp");
                    f_generico.right_temp=(double*) myalloc((n+1)*N, sizeof(double), 1, "main", "matrix_messages_right_temp");
                    #else
                    matrix_messages_left = (int*) myalloc((n+1)*N, sizeof(int), 1, "main", "matrix_messages_left");
                    matrix_messages_right = (int*) myalloc((n+1)*N, sizeof(int), 1, "main", "matrix_messages_right");
                    temp_red=(int*) myalloc(N, sizeof(int), 1, "bp_sorted", "sorter");
                    f_generico.left_temp= (int*) myalloc((n+1)*N, sizeof(int), 1, "main", "matrix_messages_left_temp");
                    f_generico.right_temp=(int*) myalloc((n+1)*N, sizeof(int), 1, "main", "matrix_messages_right_temp");
                    #endif //FLOAT
                    break;
            case 21:
            case 22:
                    #ifdef FLOAT
                    matrix_messages_left = (double*) myalloc((n+1)*N, sizeof(double), 1, "main", "matrix_messages_left");
                    matrix_messages_right = (double*) myalloc((n+1)*N, sizeof(double), 1, "main", "matrix_messages_right");
                    temp_red=(double*) myalloc(N, sizeof(double), 1, "bp_sorted", "sorter");
                    f_generico.left_temp= (double*) myalloc((n+1)*N, sizeof(double), 1, "main", "matrix_messages_left_temp");
                    f_generico.right_temp=(double*) myalloc((n+1)*N, sizeof(double), 1, "main", "matrix_messages_right_temp");
                    f_gen_app.left_temp= (double*) myalloc((n+1)*N, sizeof(double), 1, "main", "matrix_messages_left_temp");
                    f_gen_app.right_temp=(double*) myalloc((n+1)*N, sizeof(double), 1, "main", "matrix_messages_right_temp");
                    #else
                    matrix_messages_left = (int*) myalloc((n+1)*N, sizeof(int), 1, "main", "matrix_messages_left");
                    matrix_messages_right = (int*) myalloc((n+1)*N, sizeof(int), 1, "main", "matrix_messages_right");
                    temp_red=(int*) myalloc(N, sizeof(int), 1, "bp_sorted", "sorter");
                    f_generico.left_temp= (int*) myalloc((n+1)*N, sizeof(int), 1, "main", "matrix_messages_left_temp");
                    f_generico.right_temp=(int*) myalloc((n+1)*N, sizeof(int), 1, "main", "matrix_messages_right_temp");
                    f_gen_app.left_temp= (int*) myalloc((n+1)*N, sizeof(int), 1, "main", "matrix_messages_left_temp");
                    f_gen_app.right_temp=(int*) myalloc((n+1)*N, sizeof(int), 1, "main", "matrix_messages_right_temp");
                    #endif //FLOAT
                    break;

            default:printf("Please change the number of the Decoding method to use.\n");
                    exit(KO_DECODING_NOT_EXIST);
                    break;
        }

  iteration_reached = (int*) myalloc(max_point, sizeof(int),1,"main","iteration_reached");
  uncodedK = (int*) myalloc(K, sizeof(int), 1, "main", "uncodedK");
  uncodedN = (int*) myalloc(N, sizeof(int), 1, "main", "uncodedN");
  coded = (int*) myalloc(N, sizeof(int), 1, "main", "coded");
  recoded = (int*) myalloc(N, sizeof(int), 1, "main", "recoded");
  channelX = (int*) myalloc(N, sizeof(int), 1, "main", "channelX");

  FBER = (int*) myalloc(N, sizeof(int), 1, "main", "FBER");
  dec_FBER = (int *) myalloc(N, sizeof(int), 1, "main", "dec");

  crc16_coded = (uint16_t *) myalloc(N, sizeof(uint16_t), 1, "main", "crc_16");
  crc32_coded = (uint32_t *) myalloc(N, sizeof(uint32_t), 1, "main", "crc_32");
  crc16_uncoded = (uint32_t *) myalloc(N, sizeof(uint32_t), 1, "main", "crc_32");
  crc32_uncoded = (uint32_t *) myalloc(N, sizeof(uint32_t), 1, "main", "crc_32");
  crc_input_coded = (unsigned char *) myalloc(N, sizeof(unsigned char), 1, "main", "crc_input");
  crc_input_uncoded = (unsigned char *) myalloc(N, sizeof(unsigned char), 1, "main", "crc_input");

  index_g = (int*) myalloc(K, sizeof(int), 1, "main", "index_g");
  index_b = (int*) myalloc(N-K, sizeof(int), 1, "main", "index_b");
  G = (int**) myalloc(N, sizeof(int*), 0, "main", "G");
  for (i=0; i<N; i++)
    G[i] = (int*) myalloc(N, sizeof(int), 1, "main", "G[i]");
  Gt = (int**) myalloc(N, sizeof(int*), 0, "main", "Gt");
  for (i=0; i<N; i++)
    Gt[i] = (int*) myalloc(N, sizeof(int), 1, "main", "Gt[i]");
   //remap = (int*) myalloc(N/2, sizeof(int), 1, "main", "remap");
  ///remap = (int*) myalloc(N, sizeof(int), 1, "main", "remap");
  noise = (double *) myalloc(N, sizeof(double), 1, "main", "noise");
  #ifdef FLOAT
  llr_in = (double *) myalloc(N, sizeof(double), 1, "main", "llr_in");
  #else
  llr_in = (int *) myalloc(N, sizeof(int), 1, "main", "llr_in");
  #endif
  llr_ch = (double *) myalloc(N, sizeof(double), 1, "main", "llr_ch");
  dec = (int *) myalloc(N, sizeof(int), 1, "main", "dec");
  decoded = (int *) myalloc(N, sizeof(int), 1, "main", "decoded");
  decodedK = (int *) myalloc(K, sizeof(int), 1, "main", "decodedK");

  switch(decoding_type)
        {case 14: ord_circolare(&f_generico,n,N);
                  break;
         case 15: ord_bi_wave(&f_generico,n,N);
                  break;
         case 16: ord_all_on(&f_generico,n,N);
                  break;
         case 17: ord_stepped(&f_generico,n,N);
                  break;
         case 18: ord_lineare(&f_generico,n,N);
                  break;
         case 19: ord_lineare_RL(&f_generico,n,N);
                  break;
         case 20: ord_circolare_RL_single_prop(&f_generico,n,N);
                  break;
         case 21: //ord_bi_wave(&f_generico,n,N);
                  //ord_all_on(&f_generico,n,N);
                  ord_circolare_RL_single_prop(&f_generico,n,N);
                  //ord_lineare_RL(&f_generico,n,N);
                  //ord_stepped(&f_generico,n,N);
                  //ord_tetra_wave_single_prop(&f_generico,n,N);
                  //ord_bi_wave_single_prop(&f_generico,n,N);
                  ord_lineare(&f_gen_app,n-1,N);
                  break;
         case 22: //ord_bi_wave_single_prop(&f_generico,n,N);
                  ord_stepped(&f_generico,n,N);
                  //ord_circolare_RL_single_prop(&f_generico,n,N);
                  //ord_lineare_RL(&f_generico,n,N);
                  //ord_tetra_wave_single_prop(&f_generico,n,N);
                  ord_lineare(&f_generico,n,N);
                  ord_lineare(&f_gen_app,n-1,N);
                  break;
         default: break;
        }
  op_per_iter = op_count(&f_generico);
  printf("Operation per iteration = %d\n",op_per_iter);
  /// get channel configuration
  get_channels(w, N, bhatta_eps, K,BEC_AWGN_choice, stopping_rule);
  index_of_w(w, index_g, index_b, N);

  /// create G matrix
  create_g(G, n);
  transpose_g(G, Gt, N);
  //create_h(n, H);
  /// compute remap sequence
 /// get_remap(remap, N);

  /// graph representation
  #ifndef FLOAT
  if((decoding_type<5))
    {assign_vertex(n, N, matrix_destinations); //Call for function computing destinations for each node

      ///save in an output file
      sprintf(tmpbuf,"%sMATRIX_K%d.txt", oname, K);
      fp_res_matrix = myfopen(tmpbuf, "w", "main");
      save_hw_check_matrix(fp_res_matrix, matrix_destinations,N, n);
      }
   #endif // FLOAT
  /// results output file
  sprintf(tmpbuf,"%sresults_K%d.txt", oname, K);
  fp_res = myfopen(tmpbuf, "w", "main");

  if(hw_autotest_file)
  {
          sprintf(tmpbuf,"%sinput_test.txt", oname);
          fp_res_in = myfopen(tmpbuf, "w", "main");
          sprintf(tmpbuf,"%soutput_test.txt", oname);
          fp_res_out = myfopen(tmpbuf, "w", "main");
  }

    if(save_hw_initial)
  {
          sprintf(tmpbuf,"%ssetup.txt", oname);
          fp_setup = myfopen(tmpbuf, "w", "main");
  }

  int *tmp; /// index for function save_details
  if(save_details_en)
  {sprintf(tmpbuf,"%sresults_detailed_K%d.txt", oname, K);
   fp_res_det = myfopen(tmpbuf, "w", "main");
    tmp=(int*) myalloc(K, sizeof(int), 1, "main", "copy_uncodedK");
  }
  int testzero;

  modulate_data test;

   switch(QAM_number)
    {
    case -1:
    case 0:   break;
    case 16:
    case 64:
    case 256:
            test.input_data=coded;
            test.seed=&rndseedunif;
            test.update_characteristics=1;
            test.SNR=20.0;///da modificare
            test.frame_size=N;
            test.M=QAM_number;
            break;
    default:
           exit(KO_NO_MODULATION);
           return 1;
    }

    if(bicm==1)
    {
//        interleave_bcim=generateRandomInterleave(N,&rndseedunif);
    }



printf("Start!\n");

  ck_1 = clock();
  ck_delta = 0;
  ck_max = MY_CLOCK_T_MAX;


  /// frame loop
  for (curr_frame=f_num; curr_frame<=max_frame; curr_frame++)
  {
    /// build new frame
    /// good
    if (hw_check_en)/// for debugging purpose of the VHDL
    //if ((decoding_type==4)||(decoding_type==5))/// for debugging purpose of the VHDL
    {testzero=1;
     while(testzero)
        {
            for (i=0; i<K; i++)
                {
                  uncodedK[i] = Lfsr(&rndseedlfsr);
                  uncodedN[index_g[i]] = uncodedK[i];
                  if(uncodedK[i]!=0)
                    testzero=0;
                }
        }

    }
    else{
        for (i=0; i<K; i++)
            {
              uncodedK[i] = Lfsr(&rndseedlfsr);
              uncodedK[i] = 1;
              uncodedN[index_g[i]] = uncodedK[i];
            }
        }
    /// bad
    for (i=0; i<N-K; i++)
      uncodedN[index_b[i]] = 0;

    /// encoder
    encode(uncodedN, coded, Gt, N);
    if(bicm==1)
    {
//        Interleave(coded,interleave_bcim,N);
    }

    if(QAM_number==0||QAM_number==-1) /// create noise array BPSK
           {
            for (i=0; i<N; i++)
              noise[i] = Gaussian(&rndseedunif);
           }

    /// Eb/N0 loop
    for (curr_point=min_point; curr_point<max_point; curr_point++)
    {
        if(QAM_number==-1){
            for (i=0; i<N; i++)
              {

              }
        }
        if(QAM_number==0) /// create noise array BPSK
           {
              level = sqrt(0.5 / rate / pow(10.0, (ebn0_min + curr_point*ebn0_step)/10.0 ));

              /// AWGN channel with BPSK modulation and soft demodulator with saturation
              //printf("\n");
              for (i=0; i<N; i++)
              {
                modulated = 1.0 - 2.0*coded[i];
                received = modulated + level*noise[i];
                llr_ch[i] = BPSK_G * received / (level*level);
                #ifdef FLOAT
                if(max_val_int < 0)
                llr_in[i] = llr_ch[i];

                else{
                if(frac_bits>=0)
                    llr_in[i] = (double) floor(llr_ch[i] * (1 << frac_bits) + 0.5);
                else
                    llr_in[i] = (double) floor(llr_ch[i] / (1 << ((-1)*frac_bits)) + 0.5);
                if (llr_in[i] > max_val_int)
                  llr_in[i] = max_val_int;
                else if (llr_in[i] < min_val_int)
                  llr_in[i] = min_val_int;
                }
                #else
                llr_in[i] = (int) floor(llr_ch[i] * (1 << frac_bits) + 0.5);
                if (llr_in[i] > max_val_int)
                  llr_in[i] = max_val_int;
                else if (llr_in[i] < min_val_int)
                  llr_in[i] = min_val_int;
                  //printf("%f %d\t\n",llr_ch[i],llr_in[i]);
                  //getchar();
                #endif // FLOAT

              }
           }
            else
            {
                free(llr_ch);
                test.SNR=(ebn0_min + curr_point*ebn0_step);
//                CmlChannel(&test);
                llr_ch=test.output;///per uniformita i risultati sono da invertire

                for (i=0; i<N; i++)
                {
                    #ifdef FLOAT
                    if(max_val_int < 0)
                        llr_in[i] = llr_ch[i];
                    else{
                        llr_in[i] = (double) floor(llr_ch[i] * (1 << frac_bits) + 0.5);
                        if (llr_in[i] > max_val_int)
                            llr_in[i] = max_val_int;
                        else if (llr_in[i] < min_val_int)
                            llr_in[i] = min_val_int;
                    }
                    #else
                    llr_in[i] = (int) floor(-llr_ch[i] * (1 << frac_bits) + 0.5);
                    if (llr_in[i] > max_val_int)
                    llr_in[i] = max_val_int;
                    else if (llr_in[i] < min_val_int)
                    llr_in[i] = min_val_int;
                    #endif // FLOAT
                }
            }

            if(bicm==1)
            {
//                Deinterleave(llr_in,interleave_bcim,N);
            }
            //llr=fopen("llr.txt","w");
            //for(dataindex=0;dataindex<N;dataindex++){
            //    fprintf(llr,"%d\n",llr_in[dataindex]);
            //}
            //fclose(llr);
    switch(decoding_type)
        {
            #ifndef FLOAT
            case 0 :init_matrix_full(matrix_messages_left_d, matrix_messages_right_d, llr_ch, index_b, index_g, nN, N, K, DBL_MAX);
                    break;
            case 1 :init_matrix_ms(matrix_messages_left, matrix_messages_right, llr_in, index_b, index_g, nN, N, K, max_val_ext);
                    break;
            #endif // FLOAT
            case 4 :
            case 6 :
            case 9 :
            case 10:
            case 11:
            case 12:
            case 13:
            case 14:
            case 15:
            case 16:
            case 17:
            case 18:
            case 19:
            case 20:
            case 21:
            case 22:
            case 7 :init_matrix_ms(matrix_messages_left, matrix_messages_right, llr_in, index_b, index_g, nN, N, K, r_init);
                            /// aggiunto per hw check
                              if((frame_by_frame)&&(!hw_autotest_file))
                              {
                              sprintf(tmpbuf,"%sinput_point_%d_frame_%d_K%d.txt", oname,curr_point,curr_frame,K);
                              fp_res_in = myfopen(tmpbuf, "w", "main");
                              save_hw_check_llr(fp_res_in,llr_in,matrix_messages_right,N);
                              fclose(fp_res_in);
                              sprintf(tmpbuf,"%soutput_point_%d_frame_%d_K%d.txt", oname,curr_point,curr_frame,K);
                              fp_res_out = myfopen(tmpbuf, "w", "main");
                              }
                              if(hw_autotest_file)
                              {
                               save_hw_check_llr(fp_res_in,llr_in,matrix_messages_right,N);
                              }
                            ///
                    break;
            #ifndef FLOAT
            case 2 :
            case 3 :init_matrix_full2(matrix_messages_left_d, matrix_messages_right_d, llr_ch, index_b, index_g, nN, N, K, DBL_MAX);
                    break;
            case 5 :
            case 8 :init_matrix_ms_reduced(matrix_messages_left, matrix_messages_right, llr_in, index_b, index_g, nN, N, K, max_val_ext);
                            /// aggiunto per hw check
                            if((frame_by_frame)&&(!hw_autotest_file))
                              {
                              sprintf(tmpbuf,"%sinput_point_%d_frame_%d_K%d.txt", oname,curr_point,curr_frame,K);
                              fp_res_in = myfopen(tmpbuf, "w", "main");
                              save_hw_check_llr(fp_res_in,llr_in,matrix_messages_right,N);
                              fclose(fp_res_in);
                              sprintf(tmpbuf,"%soutput_point_%d_frame_%d_K%d.txt", oname,curr_point,curr_frame,K);
                              fp_res_out = myfopen(tmpbuf, "w", "main");
                              }
                              if(hw_autotest_file)
                              {
                               save_hw_check_llr(fp_res_in,llr_in,matrix_messages_right,N);
                              }
                            ///
                    break;
            #endif // FLOAT
            default:printf("Please change the number of the Decoding method to use.\n");
                    exit(KO_DECODING_NOT_EXIST);
                    break;
        }
      /// iteration loop
      /// N.B. max_iter-1
      errors = 1;

      swiching_state=0;

      for (curr_iter=0; curr_iter<max_iter; curr_iter++)
      {
        switch(decoding_type)
        {
            #ifndef FLOAT
            case 0 :/// Backward recursion
                    left_message_full(matrix_messages_left_d, matrix_messages_right_d, matrix_destinations, n, N);
                    /// Forward recursion
                    right_message_full(matrix_messages_left_d, matrix_messages_right_d, matrix_destinations, n, N);
                    break;
            case 1: /// Backward recursion
                    left_message_ms(matrix_messages_left, matrix_messages_right, matrix_destinations, n, N, min_val_ext, max_val_ext, sfact_left);
                    /// Forward recursion
                    right_message_ms(matrix_messages_left, matrix_messages_right, matrix_destinations, n, N, min_val_ext, max_val_ext, sfact_right);
                    break;
            case 2: left_right_message_full2_dx(matrix_messages_left_d, matrix_messages_right_d, matrix_destinations, n, N);
                    break;
            case 3: left_right_message_full2_sx(matrix_messages_left_d, matrix_messages_right_d, matrix_destinations, n, N);
                    break;
            case 4: left_right_message_ms_sx_v2(matrix_messages_left, matrix_messages_right, matrix_destinations, n, N, min_val_ext, max_val_ext, sfact_left);
                    ///
                    if((frame_by_frame)&&(!hw_autotest_file))
                        save_hw_check_out(fp_res_out,matrix_messages_left,matrix_messages_right, nN,N);
                    ///
                    break;
            case 5: ///collegamenti diversi
                    left_right_message_ms_reduced(matrix_messages_left, matrix_messages_right, temp_red, n, N, min_val_ext, max_val_ext, sfact_left);
                    ///
                    if((frame_by_frame)&&(!hw_autotest_file))
                        save_hw_check_out_reduced(fp_res_out,matrix_messages_left,matrix_messages_right, nN,N);
                    ///
                    break;
            case 6: ///collegamenti diversi
                    left_right_message_ms_sorted(matrix_messages_left, matrix_messages_right, temp_red, n, N, min_val_ext, max_val_ext, sfact_left);
                    ///
                    if((frame_by_frame)&&(!hw_autotest_file))
                        save_hw_check_out(fp_res_out,matrix_messages_left,matrix_messages_right, nN,N);
                    ///
                    break;
            case 7: ///collegamenti corretti
                    left_right_message_ms_sorted2(matrix_messages_left, matrix_messages_right, temp_red, n, N, min_val_ext, max_val_ext, sfact_left);
                    ///
                    if((frame_by_frame)&&(!hw_autotest_file))
                        save_hw_check_out(fp_res_out,matrix_messages_left,matrix_messages_right, nN,N);
                    ///
                    break;
            case 8: ///collegamenti corretti
                    left_right_message_ms_reduced2(matrix_messages_left, matrix_messages_right, temp_red, n, N, min_val_ext, max_val_ext, sfact_left);
                    ///
                    if((frame_by_frame)&&(!hw_autotest_file))
                        save_hw_check_out_reduced(fp_res_out,matrix_messages_left,matrix_messages_right, nN,N);
                    ///
                    break;

            case 9: ///come sorted2 ma parte dal fondo ad aggiornare
                    left_right_message_ms_sorted3(matrix_messages_left, matrix_messages_right, temp_red, n, N, min_val_ext, max_val_ext, sfact_left);
                    ///
                    if((frame_by_frame)&&(!hw_autotest_file))
                        save_hw_check_out(fp_res_out,matrix_messages_left,matrix_messages_right, nN,N);
                    ///
                    break;

            case 10:///come sorted2 e sorted3 chiamati alternativamente
                    if(swiching_state)
                        {left_right_message_ms_sorted3(matrix_messages_left, matrix_messages_right, temp_red, n, N, min_val_ext, max_val_ext, sfact_left);
                        swiching_state=0;
                        }
                    else
                        {
                        left_right_message_ms_sorted2(matrix_messages_left, matrix_messages_right, temp_red, n, N, min_val_ext, max_val_ext, sfact_left);
                        swiching_state=1;
                        }
                    ///
                    if((frame_by_frame)&&(!hw_autotest_file))
                        save_hw_check_out(fp_res_out,matrix_messages_left,matrix_messages_right, nN,N);
                    ///
                    break;

            case 11:///come sorted2 e sorted3 chiamati alternativamente
                    if(swiching_state)
                        {left_right_message_ms_sorted2(matrix_messages_left, matrix_messages_right, temp_red, n, N, min_val_ext, max_val_ext, sfact_left);
                        swiching_state=0;
                        }
                    else
                        {
                        left_right_message_ms_sorted3(matrix_messages_left, matrix_messages_right, temp_red, n, N, min_val_ext, max_val_ext, sfact_left);
                        swiching_state=1;
                        }
                    ///
                    if((frame_by_frame)&&(!hw_autotest_file))
                        save_hw_check_out(fp_res_out,matrix_messages_left,matrix_messages_right, nN,N);
                    ///
                    break;
            case 12:///parte sorted3 e poi solo sorted2///fa schifo
                    if(swiching_state)
                        {left_right_message_ms_sorted2(matrix_messages_left, matrix_messages_right, temp_red, n, N, min_val_ext, max_val_ext, sfact_left);
                        swiching_state=1;
                        }
                    else
                        {
                        left_right_message_ms_sorted3(matrix_messages_left, matrix_messages_right, temp_red, n, N, min_val_ext, max_val_ext, sfact_left);
                        swiching_state=1;
                        }
                    ///
                    if((frame_by_frame)&&(!hw_autotest_file))
                        save_hw_check_out(fp_res_out,matrix_messages_left,matrix_messages_right, nN,N);
                    ///
                    break;
            case 13:///parte sorted2 e poi solo sorted3
                    if(swiching_state)
                        {left_right_message_ms_sorted3(matrix_messages_left, matrix_messages_right, temp_red, n, N, min_val_ext, max_val_ext, sfact_left);
                        swiching_state=1;
                        }
                    else
                        {
                        left_right_message_ms_sorted2(matrix_messages_left, matrix_messages_right, temp_red, n, N, min_val_ext, max_val_ext, sfact_left);
                        swiching_state=1;
                        }
                    ///
                    if((frame_by_frame)&&(!hw_autotest_file))
                        save_hw_check_out(fp_res_out,matrix_messages_left,matrix_messages_right, nN,N);
                    ///
                    break;
            #endif // FLOAT
            case 14:
            case 15:
            case 16:
            case 17:
            case 20:left_right_message_ms_generic(matrix_messages_left, matrix_messages_right, temp_red, f_generico, n, N, min_val_ext, max_val_ext, sfact_left, &opcount);
                    ///
                    if((frame_by_frame)&&(!hw_autotest_file))
                        save_hw_check_out(fp_res_out,matrix_messages_left,matrix_messages_right, nN,N);
                    ///
                    break;
           case 19:
           case 18:left_right_message_ms_generic(matrix_messages_left, matrix_messages_right, temp_red, f_generico, n, N, min_val_ext, max_val_ext, sfact_left, &opcount);
                    ///
                    if((frame_by_frame)&&(!hw_autotest_file))
                        save_hw_check_out(fp_res_out,matrix_messages_left,matrix_messages_right, nN,N);

                    if((curr_iter==0)&&(save_hw_initial))
                        {save_hw_left_or_right(fp_setup,matrix_messages_right, n,N);
                        fclose(fp_setup);
                        }
                    ///
                    break;
           case 21:

               if(curr_iter==0)
                    {///questa parte serve a inizializzare i valori opportuni nella memoria right
/*
                    for(zz=0;zz<(n+1)*N;zz++){
                        fprintf(testfp,"%f \t\t",matrix_messages_left[zz]);
                        if(zz%32==31)
                            fprintf(testfp,"\n");
                    }
                    fprintf(testfp,"\n\n");
                    for(zz=0;zz<(n+1)*N;zz++){
                        fprintf(testfp,"%f \t\t",matrix_messages_right[zz]);
                        if(zz%32==31)
                            fprintf(testfp,"\n");
                    }
                    fprintf(testfp,"\n\n\n");
*/
                    left_right_message_ms_generic(matrix_messages_left, matrix_messages_right, temp_red, f_gen_app, n, N, min_val_ext, max_val_ext, sfact_left, &opcount);
/*
                    for(zz=0;zz<(n+1)*N;zz++){
                        fprintf(testfp,"%d \t",matrix_messages_left[zz]);
                        if(zz%8==7)
                            fprintf(testfp,"\n");
                    }
                    fprintf(testfp,"\n\n");
                    for(zz=0;zz<(n+1)*N;zz++){
                        fprintf(testfp,"%d \t",matrix_messages_right[zz]);
                        if(zz%8==7)
                            fprintf(testfp,"\n");
                    }
*/
                    //swiching_state=1;
                    //free(f_gen_app.left_temp);
                    //free(f_gen_app.right_temp);
                    //print_matrix(matrix_messages_left,n+1,N);
                    //print_matrix(matrix_messages_right,n+1,N);
                    if((curr_iter==0)&&(save_hw_initial))
                        {save_hw_left_or_right(fp_setup,matrix_messages_right, n, N);
                        fclose(fp_setup);
                        }
                    }
                    left_right_message_ms_generic(matrix_messages_left, matrix_messages_right, temp_red, f_generico, n, N, min_val_ext, max_val_ext, sfact_left, &opcount);
                    //printf("--\n");
                    //print_matrix(matrix_messages_right,n+1,N);
                    //print_matrix(matrix_messages_left,n+1,N);
                    if((frame_by_frame)&&(!hw_autotest_file))
                        save_hw_check_out(fp_res_out,matrix_messages_left,matrix_messages_right, nN,N);
/*
                    fprintf(testfp,"\n\n\n");
                    for(zz=0;zz<(n+1)*N;zz++){
                        fprintf(testfp,"%f \t\t",matrix_messages_left[zz]);
                        if(zz%32==31)
                            fprintf(testfp,"\n");
                    }
                    fprintf(testfp,"\n\n\n");
                    for(zz=0;zz<(n+1)*N;zz++){
                        fprintf(testfp,"%f  \t\t",matrix_messages_right[zz]);
                        if(zz%32==31)
                            fprintf(testfp,"\n");
                    }
*/
                    break;
            case 22: if(curr_iter==0)
                    {///questa parte serve a inizializzare i valori opportuni nella memoria right
                    left_right_message_ms_generic(matrix_messages_left, matrix_messages_right, temp_red, f_gen_app, n, N, min_val_ext, max_val_ext, sfact_left, &opcount);
                    if((curr_iter==0)&&(save_hw_initial))
                        {save_hw_left_or_right(fp_setup,matrix_messages_right, n, N);
                        fclose(fp_setup);
                        }
                    }
                    left_right_message_ms_generic(matrix_messages_left, matrix_messages_right, temp_red, f_generico, n, N, min_val_ext, max_val_ext, sfact_left, &opcount);

                    if((frame_by_frame)&&(!hw_autotest_file))
                        save_hw_check_out(fp_res_out,matrix_messages_left,matrix_messages_right, nN,N);
                    break;
            default:printf("Please change the number of the Decoding method to use.\n");
                    exit(KO_DECODING_NOT_EXIST);
                    break;

        }

        /// hard decision
        switch(decoding_type)
        {
            #ifndef FLOAT
            case 2 :
            case 3 :
            case 0 :for (i=0; i<N; i++)
                    {
                     if (matrix_messages_left_d[i]> 0.0)
                        dec[i] = 0;
                      else
                        dec[i] = 1;
                    }
                    decoded=dec;
                    break;
            case 4 :
            case 5 :
            case 6 :
            case 7 :
            case 8 :
            case 9 :
            case 10:
            case 11:
            case 12:
            case 13:
            #endif // FLOAT
            case 14:
            case 15:
            case 16:
            case 17:
            case 18:
            case 19:
            case 20:
            case 21:
            case 22:
            case 1 :

            for (i=0; i<N; i++)
                {
                    if (matrix_messages_left[i] >= 0)
                        dec[i] = 0;
                    else
                        dec[i] = 1;
                }
                    //encode(dec, decoded, Gt, N);
            decoded=dec;
            for (i=0; i<N; i++)
            {
                if((matrix_messages_left[nN+i] + matrix_messages_right[nN+i]) >=0)
                    channelX[i]=0;
                else
                    channelX[i]=1;
            }
            break;
            default:printf("Please change the number of the Decoding method to use.\n");
                    exit(KO_DECODING_NOT_EXIST);
                    break;
        }



    for (i=0; i<K; i++)
        decodedK[i]=decoded[index_g[i]];

	/// statistics update
	avg_iter[curr_point]++;
	errors = BER_meter(K, decodedK, uncodedK, bit_err, frame_err, curr_iter, curr_point);

//	if(((decoding_type==4)||(decoding_type==5))&&(!errors)&&(curr_iter>1))/// //////////////////////////////////aggiunto
//    {
//        printf("stop");
//       // fclose(fp_res_out);
//    }

    /// Stopping rules: 0 none, 1 no errori, 2 frozen bit, 3 Gmatrix, 5 FBER
     exit_ite_loop=0;
     M=4;
     switch(stopping_rule)
        {
                case 5:
                    FBER_temp=0;
                    for (i=0; i<N; i++)
                        {
                            if (matrix_messages_left[i] >= 0) dec_FBER[i] = 0;
                            else dec_FBER[i] = 1;
                            FBER_temp+=(((1+dec_FBER[i])/2)^coded[i]);
                            FBER[curr_iter]=FBER_temp;
                        }
                    if (curr_iter>0)
                        {
                            FBER_M=FBER[curr_iter+1]-FBER[curr_iter];
                            if (FBER_M==0) FBER_count++;
                            else FBER_count=0;
                            if (FBER_count==M)
                            {
                                exit_ite_loop=1;
                                FBER_count=0;
                                FBER_stat++;
                            }
                        }
                    if (curr_iter==max_iter) FBER_M=FBER_count=0;
                    break;
            case 3:
                  for (i=0; i<N-K; i++)
                    decoded[index_b[i]]=0;
                  encode(decoded, recoded, Gt, N);
                  error_stop=0;
                  for(i=0; i<N; i++){
                    error_stop += recoded[i] ^ channelX[i];
                  }
                  if(error_stop==0)
                  {
                      exit_ite_loop=1;
                      G_stat++;
                  }

                  break;
            case 2:
                  continue_decoding=0;
                  for (i=0; i<(N-K); i++)
                    {if(decoded[index_b[i]]==0)
                        {
                         continue_decoding++;
                        }
        }
                 if(continue_decoding==(N-K))
                        exit_ite_loop=1;
                    break;
            case 1:
                        if(!errors)
                          {
                           exit_ite_loop=1;
                          }
                        break;
            case 0:
            default:   break;
        }

        ///CRC
        if(exit_ite_loop==0)
            {
                crc_case=32;
                if (curr_iter==0) for(i=0;i<N;i++) crc_input_uncoded[i]=uncodedN[i];
                for(i=0;i<N;i++) crc_input_coded[i]=decoded[i];
                switch(crc_case)
                    {
                    case 32:
                        if (curr_iter==0) crc32_uncoded=crc_32(crc_input_uncoded,N);
                        crc32_coded[curr_iter]=crc_32(crc_input_coded,N);
                        if (curr_iter>0 && crc32_coded[curr_iter]==crc32_uncoded) crc_count++;
                        else crc_count=0;
                        break;
                    case 16:
                        if (curr_iter==0) crc16_uncoded=crc_16(crc_input_uncoded,N);
                        crc16_coded[curr_iter]=crc_16(crc_input_coded,N);
                        if (curr_iter>0 && crc16_coded[curr_iter]==crc16_uncoded) crc_count++;
                        else crc_count=0;
                        break;
                    case 0:
                        default:  crc_count=0;
                        break;
                    }
                    if (crc_count==1)
                        {
                        exit_ite_loop=1;
                        crc_count=0;
                        crc_stat++;
                        }
            if (curr_iter==max_iter) crc_count=0;
            }
    if(exit_ite_loop==1)
        break;


    ///STOPPING STATISTICS
    tot_stop_frame=G_stat+FBER_stat+crc_stat;
    G_stat_per=(G_stat*100)/(double)tot_stop_frame;
    FBER_stat_per=(FBER_stat*100)/(double)tot_stop_frame;
    crc_stat_per=(crc_stat*100)/(double)tot_stop_frame;

    } /// end iteration loop

      if(hw_autotest_file)
                  {if((decoding_type==5)||(decoding_type==8))
                   save_hw_check_out_reduced(fp_res_out,matrix_messages_left,matrix_messages_right, nN,N);
                   else
                   save_hw_check_out(fp_res_out,matrix_messages_left,matrix_messages_right, nN,N);
                  }

      if(frame_by_frame)
        fclose(fp_res_out);


    ///save the decoded frame and the number of used iterations
      if(save_details_en)
      save_details(fp_res_det,curr_iter,K, decodedK, uncodedK,curr_point,max_point,ebn0_min,ebn0_step,curr_frame,tmp,min_point);

      /// update the maximum point computed so far
      if(curr_point > max_point_comp)
	   max_point_comp = curr_point;
      /// Check if no errors have been detected in the minimal iteration
//      if(curr_iter < min_iter)
//        break;

    } //for (curr_point=min_point; curr_point<max_point; curr_point++)

    /// update the total number of frames
    for (curr_point=min_point; curr_point<max_point; curr_point++)
    {
      tot_frames[curr_point]++;
      /// Check if a reliable point is ready
      if( frame_err[max_iter-1][min_point] >= max_err){
        iteration_reached[min_point] = curr_frame;
        min_point++;
      }

    }

    /// Check if all point are reliable
    //if(min_point == max_point)
      //break;

    ck_2 = clock();

    /// check the elapsed time and save results
    if (ck_2 > ck_1)
      ck_delta = ck_2 - ck_1;
    else
      ck_delta = ck_max - ck_1 + ck_2;

    if( ck_delta/CLOCKS_PER_SEC >= time_check )
    {
      store_results(fp_res, min_iter, max_iter, max_point_comp, ebn0_min, ebn0_step, bit_err, frame_err, tot_frames, K, avg_iter, iteration_reached);
      save_simulation(sim_par, lfsr_seed, unif_seed, curr_frame, max_point_comp, min_point, tot_frames, frame_err, bit_err, avg_iter);
      if(stopping_rule!=0 || crc_case!=0)
        {
            printf("Stopped frames=%d, G-MATRIX=%d (%2.3f%), FBER=%d (%2.3f%), CRC=%d (%2.3f%)\n",tot_stop_frame+1,G_stat,G_stat_per,FBER_stat,FBER_stat_per,crc_stat,crc_stat_per);
            printf("frame %d of %d. Actual minimum point %d\n", curr_frame, max_frame, min_point);
        }
     //aggiunto
      acc_time+=(float)(ck_delta/CLOCKS_PER_SEC);
      velocity=((float)curr_frame)/acc_time;
      printf("estimated time left: %f min\n",((float)(max_frame-curr_frame))/velocity/60);
      //
      ck_1=ck_2;
    }

  } //for (curr_frame=f_num; curr_frame<=max_frame; f++)

  store_results(fp_res, min_iter, max_iter, max_point_comp, ebn0_min, ebn0_step, bit_err, frame_err, tot_frames, K, avg_iter, iteration_reached);
  save_simulation(sim_par, lfsr_seed, unif_seed, curr_frame, max_point_comp, min_point, tot_frames, frame_err, bit_err, avg_iter);

  fprintf(fp_res,"\n");
  fprintf(fp_res,"********* Operation per Iteration **********\n");
  for (i=0; i<max_point;i++){
    fprintf(fp_res,"Each iteration requires %d basic operation\n",op_per_iter);
  }
  fprintf(fp_res,"\n");
  fprintf(fp_res,"********* Stopping statistics **********\n");
  fprintf(fp_res,"Stopped iterations=%d\n",tot_stop_frame);
  fprintf(fp_res,"G-MATRIX=%d (%2.3f%)\n",G_stat,G_stat_per);
  fprintf(fp_res,"FBER=%d (%2.3f%)\n",FBER_stat,FBER_stat_per);
  fprintf(fp_res,"CRC=%d (%2.3f%)\n",crc_stat,crc_stat_per);


  fclose(fp_res);
  if(save_details_en)
    fclose(fp_res_det);

  //fclose(fp_res_out);

  if(decoding_type==5)
    free(temp_red);
  //fclose(testfp);
  return OK_RET;
}
