#include<stdio.h>
#include<stdlib.h>
#include<string.h>

#include"myretcodes.h"
#include"myservs.h"

/// my wrap to fopen with pointer check
///\param *fname is the file name
///\param *mod is the file opening mode "r", "w", ...
///\param *fcaller is a string containing the name of the calling function
FILE* myfopen(char *fname, char *mod, char *fcaller)
{
  FILE *fp;

  fp = fopen(fname, mod);
  if (fp == NULL)
  {
    printf("Error in %s: cannot open %s in %s\n", fcaller, fname, mod);
    exit(KO_FOPEN_NULL);
  }

  return fp;
}

/// my wrap to malloc/calloc with pointer check
///\param nelem is the number of element to allocate
///\param nsize is the size of each element
///\param init is a boolean value: 1 memory is initialized to zero; 0 memory is not itialized
///\param *fcaller is a string containing the name of the calling function
///\param *aname is a string containing the name of the array that will get allocation pointer
void* myalloc(int nelem, int nsize, int init, char *fcaller, char *aname)
{
  void *ptr;

  if (init == 1)
    ptr = calloc(nelem, nsize);
  else
    ptr = malloc(nelem*nsize);
  
  if (ptr == NULL)
  {
    printf("Error in %s: cannot allocate memory for %s\n", fcaller, aname);
    exit(KO_ALLOC_NULL);
  }

  return ptr;
}

/// my wrap to fgets with fgets return value check
///\param *buf is the buffer read from file *fp
///\param N is the maximum number of characters read from *fp (including newline)
///\param *fp is the file pointer
///\param *cname is the calling function name
///\param *dstring is a description string
void myfgets(char *buf, int N, FILE *fp, const char *cname, const char *dstring)
{
  char *ptr;

  ptr = fgets(buf, N, fp);
  if (ptr == NULL)
  {
    printf("Error: in %s fgets returned NULL while trying to read %s\n", cname, dstring);
    exit(KO_FGETS_NULL);
  }
}

/// check the fscanf return value, namely compared val and N
///\param val is the fscanf return value
///\param N is the number of data to read in fscanf
///\param dstring is a description string
void tmp_check(int val, int N, const char *dstring)
{
  if (val != N)
  {
    printf("Error:  %d elements out of %d fscanf read for %s\n", val, N, dstring);
    exit(KO_FSCANF);
  }
}


