#include<stdio.h>
#include<stdlib.h>
#include"params.h"
#include"myservs.h"

/// input parameter function: reads from a file the configuration for current simulation
///\param *ptr is the pointer to the parameter structure
void get_params(FILE *fp, params_t *ptr)
{
  char stringa[LARGE_INT_VAL];
  int tmp;
  int i,j;

  /************************ Output file **********************************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "oname line comment");
  tmp=fscanf(fp, "%s\n", ptr->oname);
  tmp_check(tmp, 1, "oname");
  /************************ Check results period **********************************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "time_check line comment");
  tmp=fscanf(fp, "%d\n", &(ptr->time_check));
  tmp_check(tmp, 1, "time_check");
  /*************************** uncoded frame size *************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "K line comment");
  tmp=fscanf(fp, "%d\n", &(ptr->K));
  tmp_check(tmp, 1, "K");
  /*************************** coded frame size *************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "N line comment");
  tmp=fscanf(fp, "%d\n", &(ptr->N));
  tmp_check(tmp, 1, "N");
  /************************** Bhattacharyya epsilon *****************************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "Bhattacharyya epsilon line comment");
  tmp=fscanf(fp,"%lf\n", &(ptr->bhatta_eps));
  tmp_check(tmp, 1, "bhatta_eps");
  /*************************** Max number of iterations *************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "min_iter, max_iter line comment");
  tmp=fscanf(fp, "%d,%d\n", &(ptr->min_iter), &(ptr->max_iter));
  tmp_check(tmp, 2, "min_iter, max_iter");
  /**************************  Max number of frame simulated *************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "max_frame line comment");
  tmp=fscanf(fp, "%d\n", &(ptr->max_frame));
  tmp_check(tmp, 1, "max_frame");
  /**************************  Max number of wrong frames simulated *************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "max_err line comment");
  tmp=fscanf(fp, "%d\n", &(ptr->max_err));
  tmp_check(tmp, 1, "max_err");
  /************************** Eb/N0 *****************************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "ebn0_min, ebn0_step, max_point line comment");
  tmp=fscanf(fp,"%lf,%lf,%d\n", &(ptr->ebn0_min), &(ptr->ebn0_step), &(ptr->max_point));
  tmp_check(tmp, 3, "ebn0_min, ebn0_step, max_point");
  /************************** lfsr_seed *****************************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "lfsr_seed line comment");
  tmp=fscanf(fp,"%d, %d\n", &(ptr->lfsr_seed),&(ptr->lfsr_sel));
  tmp_check(tmp, 2, "lfsr_seed");
  /************************** unif_seed *****************************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "unif_seed line comment");
  tmp=fscanf(fp,"%d, %d\n", &(ptr->unif_seed),&(ptr->unif_sel));
  tmp_check(tmp, 2, "unif_seed");
  /*********************** intrinsic information fractional bits, minimum and maximum value **************************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "frac bits, int. inf. min and max line comment");
  tmp=fscanf(fp,"%d,%d,%d\n", &(ptr->frac_bits), &(ptr->min_val_int), &(ptr->max_val_int));
  tmp_check(tmp, 3, "frac_bits, min_val_int, max_val_int");
  /*********************** extrinsic information minimum and maximum value **************************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "ext. inf. min and max line comment");
  tmp=fscanf(fp,"%d,%d\n", &(ptr->min_val_ext), &(ptr->max_val_ext));
  tmp_check(tmp, 2, "min_val_ext, max_val_ext");
  /*********************** extrinsic information minimum and maximum value **************************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "R init value");
  tmp=fscanf(fp,"%d\n", &(ptr->R_init));
  tmp_check(tmp, 1, "r_init");
  /*********************** scaling factor for min-sum algorithm **************************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "scaling factor line comment");
  tmp=fscanf(fp,"%f,%f\n", &(ptr->sfact_left), &(ptr->sfact_right));
  tmp_check(tmp, 2, "scaling factors");
  /*********************** Starting frame number (for resuming simulation) ***************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "starting frame number line comment");
  tmp=fscanf(fp,"%d\n", &(ptr->f_num));
  tmp_check(tmp, 1, "starting frame number");
  /*********************** Decoding type to use (number) *********************************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "decoding type number line comment");
  tmp=fscanf(fp,"%d\n", &(ptr->decoding_type_number));
  tmp_check(tmp, 1, "decoding type number");
  /*********************** QAM modulation to use (number) *********************************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "QAM modulation and BICM number line comment");
  tmp=fscanf(fp,"%d, %d\n", &(ptr->QAM_modulation_number), &(ptr->BICM_enable));
  tmp_check(tmp, 2, "QAM BICM type number");
  /*********************** Stopping rules *********************************/
  myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "Stopping line comment");
  tmp=fscanf(fp, "%d\n", &(ptr->stopping));
  tmp_check(tmp, 1, "Stopping rule number");

  /// initialization and allocation
  ptr->max_point_comp = 0;
  ptr->min_point = 0;

  ptr->tot_frames = (int*) myalloc(ptr->max_point, sizeof(int), 1, "get_params", "tot_frames");
  ptr->frame_err = (int**) myalloc(ptr->max_iter, sizeof(int*), 0, "get_params", "frame_err");
  for (i=0; i<ptr->max_iter; i++)
    (ptr->frame_err)[i] = (int*) myalloc(ptr->max_point, sizeof(int), 1, "get_params", "frame_err[i]");
  ptr->bit_err = (int**) myalloc(ptr->max_iter, sizeof(int*), 0, "get_params", "bit_err");
  for (i=0; i<ptr->max_iter; i++)
    (ptr->bit_err)[i] = (int*) myalloc(ptr->max_point, sizeof(int), 1, "get_params", "bit_err[i]");
  ptr->avg_iter = (int*) myalloc(ptr->max_point, sizeof(int), 1, "get_params", "avg_iter");

  /// check f to resume simulation
  if (ptr->f_num > 1)
  {
    printf("Resuming simulation\n");

    /***************** Maximum point computed ************/
    myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "max_point_comp line comment");
    tmp=fscanf(fp,"%d\n", &(ptr->max_point_comp));
    tmp_check(tmp, 1, "max_point_comp");
    /************************** Minimum point to compute ************/
    myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "min_point line comment");
    tmp = fscanf(fp,"%d\n", &(ptr->min_point));
    tmp_check(tmp, 1, "min_point");
    /***** Retrieve status of counters */
    myfgets(stringa, LARGE_INT_VAL, fp, "get_params", "counter status line comment");
    for(j=0; j<ptr->max_point; j++)
    {
      tmp = fscanf(fp,"%d\n", &((ptr->tot_frames)[j]));
      for(i=0; i<ptr->max_iter; i++)
	tmp = fscanf(fp, "%d,", &((ptr->bit_err)[i][j]));
      tmp = fscanf(fp,"\n");
      for(i=0; i<ptr->max_iter; i++)
	tmp = fscanf(fp,"%d,",&((ptr->frame_err)[i][j]));
      tmp = fscanf(fp,"\n");
      tmp = fscanf(fp,"%d\n", &((ptr->avg_iter)[j]));
    }
  }
}


/// show parameter function: print on screen the configuration for current simulation
///\param sim_par is the parameter structure
///\param rate is the code rate (K/N)
void show_params(params_t sim_par, double rate)
{
  printf("Simulation parameters:\n");
  printf("K=%d, N=%d, rate=%lf, Bhattacharyya epsilon=%lf\n", sim_par.K, sim_par.N, rate, sim_par.bhatta_eps);
  printf("Using %d bits for fractional representation\n", sim_par.frac_bits);
  printf("Intrinsic information: from %d to %d\n", sim_par.min_val_int, sim_par.max_val_int);
  printf("Extrinsic information: from %d to %d\n", sim_par.min_val_ext, sim_par.max_val_ext);
  printf("R value initialization: %d\n", sim_par.R_init);
  printf("Simulating %d frames and max %d wrong frames\n", sim_par.max_frame, sim_par.max_err);
  printf("Simulating %d points from %lf with step %lf\n", sim_par.max_point, sim_par.ebn0_min, sim_par.ebn0_step);
  printf("Simulating %d - %d iterations\n", sim_par.min_iter, sim_par.max_iter);
  printf("Pseudo-random generator seeds: lfsr=%u type: %d; unif=%u type: %d\n", sim_par.lfsr_seed,sim_par.lfsr_sel, sim_par.unif_seed, sim_par.unif_sel);
  printf("Simulation checked every %d seconds\n", sim_par.time_check);
  printf("Decoding type number %d used.\n", sim_par.decoding_type_number);
  if(sim_par.decoding_type_number==0||sim_par.decoding_type_number==2)
    printf("No extrinsic and intrinsic representation limits.\n");
  printf("Stopping rule %d used.\n", sim_par.stopping);
  if(sim_par.QAM_modulation_number==0)
    printf("BPSK data modulation used.\n");
  else
    printf("%d-QAM data modulation used.\n",sim_par.QAM_modulation_number);
  if(sim_par.BICM_enable==0)
    printf("NO BICM used.\n");
  else
    printf("BICM used.\n");
}

