#ifndef _PARAMS_H
#define _PARAMS_H

#include"myservs.h"

typedef struct params_struct {
  int K; /// uncoded size
  int N; /// coded size

  int max_frame; /// maximum number of frames
  int max_err; /// maximum number of wrong frames
  double ebn0_min; /// first SNR value
  double ebn0_step; /// SNR simulation step
  int max_point; /// maximum number of SNR points
  int min_point; /// minimum number of SNR points
  int max_point_comp; /// maximum SNR point computed

  int min_iter; /// minimum number of iterations
  int max_iter; /// maximum number of iterations

  int max_val_int; /// intrinsic information maximum value (for saturation)
  int min_val_int; /// intrinsic information minimum value (for saturation)
  int frac_bits; /// number of fractional bit for LLR generation
  int min_val_ext; /// minimum value for extrinsic information
  int max_val_ext; /// maximum value for extrinsic information

  int R_init;///R messages initialization value

  float sfact_left; /// scaling factor for min-sum algorithm left messages
  float sfact_right; /// scaling factor for min-sum algorithm right messages

  char oname[LARGE_INT_VAL]; /// output result file name
  int time_check; /// simulation time check for save/resume purpose
  unsigned int lfsr_seed; /// seed for lfsr (data) random generator
  unsigned int lfsr_sel;
  unsigned int unif_seed; /// uniform generator seed (for AWGN)
  unsigned int unif_sel;

  int f_num; /// starting frame number

  int *tot_frames; /// total number of frames counter
  int **bit_err; /// error bit counter
  int **frame_err; /// errror frame counter
  int *avg_iter; /// average number of iterations

  double bhatta_eps; /// Bhattacharyya epsilon

  int decoding_type_number; /// number of the decoding to use
  int QAM_modulation_number;/// 0->PBSK, 16,64,256-QAM
  int BICM_enable;
  int stopping;

} params_t;

/// input parameter function: reads from a file the configuration for current simulation
///\param *ptr is the pointer to the parameter structure
void get_params(FILE *fp, params_t *ptr);

/// show parameter function: print on screen the configuration for current simulation
///\param sim_par is the parameter structure
///\param rate is the code rate (K/N)
void show_params(params_t sim_par, double rate);

#endif
