#include<stdio.h>
#include<stdlib.h>

#include"results.h"
#include"myservs.h"
#include"myretcodes.h"

//#define DISPLAY_BER

/// measure the ber and the fer
///\param K is the block size
///\param *dec is the hard decision couple
///\param *frame is the original uncoded frame
///\param **bit_err is the array where bit error counters are stored
///\param **frame_err is the array where frame error counters are stored
///\param itc is the current iteration
///\param curr_point is the current SNR point
///\return the number of error counted for current frame
int BER_meter(int K, int *dec, int *frame, int **bit_err, int **frame_err, int itc, int curr_point)
{
  int i;
  int count;

  /// error counter
  count=0;
  for (i=0; i<K; i++)
  {
     if (dec[i] != frame[i])
        count++;
    //count += dec[i] ^ frame[i];
  }

  bit_err[itc][curr_point] += count;
  if (count != 0)
    frame_err[itc][curr_point]++;

//  bit_err[itc] += count;
//  if (count != 0)
//    frame_err[itc]++;

  return count;
}

/// stores the simulation results (ber/fer)
///\param *fp is the results file pointer
///\param min_iter is the minimun number of iterations
///\param max_iter is the maximum number of iterations
///\param max_point_comp is the maximum SNR point computed so far
///\param ebn0_min is the minimum SNR value simulated
///\param ebn0_step is the SNR step between consecutive SNR points
///\param **bit_err is the array where bit error counters are stored
///\param **frame_err is the array where frame error counters are stored
///\param *tot_frames is the array where the total number of frames simulated per each SNR point is stored
///\param K is the frame size
///\param *avg_iter is the average number of iterations
void store_results(FILE *fp, int min_iter, int max_iter, int max_point_comp, double ebn0_min, double ebn0_step, int **bit_err, int **frame_err, int *tot_frames, int K, int *avg_iter, int *itReach)
{
  int i,j;

  rewind(fp);

  fprintf(fp,"********* BER Results **********\n");
  fprintf(fp,"       ");
  for(i=min_iter-1; i<max_iter; i++)
  {
#ifdef DISPLAY_BER
    printf("%8d|", i+1);
#endif
    fprintf(fp, "%8d|", i+1);
  }
#ifdef DISPLAY_BER
  printf("\n");
#endif
  fprintf(fp,"\n");

  for(j=0; j<max_point_comp+1; j++)
  {
#ifdef DISPLAY_BER
    printf("%6.2f|", ebn0_min+j*ebn0_step);
#endif
    fprintf(fp, "%6.2f|", ebn0_min+j*ebn0_step);
    for(i=min_iter-1; i<max_iter; i++)
    {
#ifdef DISPLAY_BER
      printf("%.2e|", (bit_err[i][j])/(double)tot_frames[j]/K);
#endif
      fprintf(fp, "%.2e|", ((double)bit_err[i][j])/(double)tot_frames[j]/(double)K);
    }
#ifdef DISPLAY_BER
    printf("\n");
#endif
    fprintf(fp,"\n");
  }

  fprintf(fp,"\n");
  fprintf(fp,"********* FER Results **********\n");
  fprintf(fp,"       ");
  for(i=min_iter-1; i<max_iter; i++)
  {
    fprintf(fp,"%8d|", i+1);
  }
  fprintf(fp,"\n");
  for(j=0; j<max_point_comp+1; j++)
  {
    fprintf(fp, "%6.2f|", ebn0_min+j*ebn0_step);
    for(i=min_iter-1; i<max_iter; i++)
    {
      fprintf(fp, "%.2e|", ((double)frame_err[i][j])/(double)tot_frames[j]);
    }
    fprintf(fp,"\n");
  }

  fprintf(fp,"\n");
  fprintf(fp,"********* Iterations **********\n");
  for(j=0; j<max_point_comp+1; j++)
  {
    fprintf(fp, "%6.2f\t", ebn0_min+j*ebn0_step);
    fprintf(fp, "%.2e\n", (avg_iter[j])/(double)tot_frames[j]);
  }

  fprintf(fp,"\n");
  fprintf(fp,"********* Error detected **********\n");
  for(j=0; j<max_point_comp+1; j++)
  {
    fprintf(fp, "%6.2f\t", ebn0_min+j*ebn0_step);
    fprintf(fp, "%d\n", frame_err[max_iter-1][j]);
  }

  fprintf(fp,"\n");
  fprintf(fp,"********* Iteration Reached **********\n");
  for(j=0; j<max_point_comp+1; j++)
  {
    fprintf(fp, "%6.2f\t", ebn0_min+j*ebn0_step);
    fprintf(fp, "%d\n", itReach[j]);
  }
  fflush(fp);

  printf("Results stored.\n");
}

/// save simulation status for future resuming
///\param lfsr_seed is the seed of the lfsr used to randomly generate uncoded bits
///\param unif_seed is the seed of the uniform generator used to produce gaussian noise samples
///\param curr_frame is the current frame number (for resuming simulation)
///\param max_point_comp is the maximum simulation point computed so far
///\param min_point minimum current starting point
///\param *tot_frames is the total number of simulated frame array
///\param **frame_err is the frame error counter array
///\param **bit_err is the bit error counter array
///\param *avg_iter is the average iteration counter array
void save_simulation(params_t sim_par, unsigned int lfsr_seed, unsigned int unif_seed, int curr_frame, int max_point_comp, int min_point, int *tot_frames, int **frame_err, int **bit_err, int *avg_iter)

{
  FILE *fp;
  int i,j;

  fp = myfopen("resumed.txt", "w", "save_simulation");

  /************************ Output file **********************************/
  fprintf(fp, "Output file\n");
  fprintf(fp, "%s\n", sim_par.oname);
  /************************ Check results period **********************************/
  fprintf(fp, "Check results period\n");
  fprintf(fp, "%d\n", sim_par.time_check);
  /*************************** uncoded frame size *************/
  fprintf(fp, "Uncoded frame size\n");
  fprintf(fp, "%d\n", sim_par.K);
  /*************************** coded frame size *************/
  fprintf(fp, "Coded frame size\n");
  fprintf(fp, "%d\n", sim_par.N);
 /************************** Bhattacharyya epsilon *****************************/
  fprintf(fp, "Bhattacharyya epsilon\n");
  fprintf(fp, "%lf\n", sim_par.bhatta_eps);
  /*************************** Max number of iterations *************/
  fprintf(fp, "Minimum-Maximum number of iterations\n");
  fprintf(fp, "%d,%d\n", sim_par.min_iter, sim_par.max_iter);
  /**************************  Max number of frame simulated *************/
  fprintf(fp, "Maximum number of frame simulated\n");
  fprintf(fp, "%d\n", sim_par.max_frame);
  /**************************  Max number of wrong frames simulated *************/
  fprintf(fp, "Maximum number of wrong frames simulated\n");
  fprintf(fp, "%d\n", sim_par.max_err);
  /************************** Eb/N0 *****************************/
  fprintf(fp, "Eb/N0: min, step, max_point\n");
  fprintf(fp,"%lf,%lf,%d\n", sim_par.ebn0_min, sim_par.ebn0_step, sim_par.max_point);
  /************************** lfsr_seed *****************************/
  fprintf(fp, "LFSR seed\n");
  fprintf(fp,"%u\n", lfsr_seed);
  /************************** unif_seed *****************************/
  fprintf(fp, "Uniform generator seed\n");
  fprintf(fp,"%u\n", unif_seed);
  /*********************** intrinsic information fractional bits, minimum and maximum value **************************/
  fprintf(fp, "frac bits, int. inf. min and max\n");
  fprintf(fp,"%d,%d,%d\n", sim_par.frac_bits, sim_par.min_val_int, sim_par.max_val_int);
  /*********************** extrinsic information minimum and maximum value **************************/
  fprintf(fp, "ext. inf. min and max\n");
  fprintf(fp,"%d,%d\n", sim_par.min_val_ext, sim_par.max_val_ext);
  /*********************** scaling factors for min-sum algorithm **************************/
  fprintf(fp, "scaling factors\n");
  fprintf(fp,"%f,%f\n", sim_par.sfact_left, sim_par.sfact_right);
  /*********************** Starting frame number (for resuming simulation) ***************/
  fprintf(fp, "starting frame number\n");
  fprintf(fp,"%d\n", curr_frame);
  /***************** Maximum point computed ************/
  fprintf(fp, "Maximum point computed\n");
  fprintf(fp,"%d\n", max_point_comp);
  /************************** Minimum point to compute ************/
  fprintf(fp, "Minimum point to compute\n");
  fprintf(fp,"%d\n", min_point);
  /***** Status of counters */
  fprintf(fp, "Status of counters\n");
  for(j=0; j<sim_par.max_point; j++)
  {
    fprintf(fp, "%d\n", tot_frames[j]);
    for(i=0; i<sim_par.max_iter; i++)
      fprintf(fp, "%d,", bit_err[i][j]);
    fprintf(fp,"\n");
    for(i=0; i<sim_par.max_iter; i++)
      fprintf(fp,"%d,", frame_err[i][j]);
    fprintf(fp,"\n");
    fprintf(fp, "%d\n", avg_iter[j]);
  }
  fclose(fp);
}


void save_details(FILE *fp,int curr_iter,int K,int *decodedK,int *uncodedK, int curr_point, int max_point, double ebn0_min, double ebn0_step,int curr_frame,int *tmp, int min_point)
{ int j,i;
if((curr_frame==1)&&(curr_point==0))
    {
        rewind(fp);
        fprintf(fp,"TX");
        for(j=0;j<K-2;j++)
            fprintf(fp," ");
        for(j=0; j<max_point; j++)
        {
        fprintf(fp, "|\t%6.2f", ebn0_min+j*ebn0_step);
        }
    fprintf(fp,"|\n");
    }
if(curr_point==min_point)
    {fprintf(fp,"\n");
     for(j=0;j<K;j++)
        {
         fprintf(fp,"%d",uncodedK[j]);
         tmp[j]=uncodedK[j];
        }
    fprintf(fp,"|\t");
    for(j=0;j<min_point;j++)
        {
         fprintf(fp,"X|X");
         for(i=0;i<K-1;i++)
             {fprintf(fp," ");}
         fprintf(fp,"|\t");
        }
    }
fprintf(fp,"%d|",curr_iter);
    for(j=0;j<K;j++)
        {fprintf(fp,"%d",decodedK[j]);
        }
    fprintf(fp,"|\t");

fflush(fp);

    ///check desynchronization (written for safety)
    for(j=0;j<K;j++)
        {if(tmp[j]!=uncodedK[j])
            {
                printf("Error: Desynchronization- frame sent ");
                for(i=0;i<K;i++)
                        {
                         printf("%d",uncodedK[i]);
                        }
                printf(" but expected to be ");
                for(i=0;i<K;i++)
                        {
                         printf("%d",tmp[i]);
                        }
                printf(".\n curr_point=%d, curr_frame=%d.\n",curr_point,curr_frame);
                exit(KO_DESYNC_SAVE_DETAIL);
            }
        }

}

#ifdef FLOAT
void save_hw_check_out_reduced(FILE *fp,double *left, double *right,int nN,int N)
#else
void save_hw_check_out_reduced(FILE *fp,int *left, int *right,int nN,int N)
#endif // FLOAT
{
        int i;

    for (i=0; i<N; i++)
      {
        #ifdef FLOAT
        fprintf(fp,"%f %f\n",left[i],right[N+i]);
        #else
        fprintf(fp,"%d %d\n",left[i],right[N+i]);
        #endif
      }
    fprintf(fp,"\n");
    fflush(fp);
}

#ifdef FLOAT
void save_hw_check_out(FILE *fp,double *left, double *right,int nN,int N)
#else
void save_hw_check_out(FILE *fp,int *left, int *right,int nN,int N)
#endif
{
        int i;

    for (i=0; i<N; i++)
      {
        #ifdef FLOAT
        fprintf(fp,"%f %f\n",left[i],right[nN+i]);
        #else
        fprintf(fp,"%d %d\n",left[i],right[nN+i]);
        #endif
      }
    fprintf(fp,"\n");
    fflush(fp);
}

#ifdef FLOAT
void save_hw_left_or_right(FILE *fp,double *left, int n,int N)
#else
void save_hw_left_or_right(FILE *fp,int *left, int n,int N)
#endif
{
    int i,j,k;

    k=0;
    fprintf(fp, "(" );

    for(j=0; j<n; j++)
    {fprintf(fp,"(");
      for (i=0; i<N; i++)
          {
              #ifdef FLOAT
              if(i==N-1)
               fprintf(fp," %f",left[k]);
              else
               fprintf(fp," %f,",left[k]);
              #else
              if(i==N-1)
               fprintf(fp," %d",left[k]);
              else
               fprintf(fp," %d,",left[k]);
              #endif // FLOAT
            k++;
          }
     if(j==n-1)
      fprintf(fp,")\n");
     else
      fprintf(fp,"),\n");
    }
    fprintf(fp,");\n");
    fflush(fp);
}

#ifdef FLOAT
void save_hw_check_llr(FILE *fp, double *llr_in,double *right,int N)
#else
void save_hw_check_llr(FILE *fp, int *llr_in,int *right,int N)
#endif
{
    int i;

    for (i=0; i<N; i++)
      {
        #ifdef FLOAT
        fprintf(fp,"%f %f\n",right[i],llr_in[i]);
        #else
        fprintf(fp,"%d %d\n",right[i],llr_in[i]);
        #endif
      }
      fprintf(fp,"\n");
      fflush(fp);
}


void save_hw_check_matrix(FILE *fp, int **matrix_destinations,int N, int n)
{
   int i,j,k;

    for (i=0; i<n; i++)
  {
    k = i*N;
    for (j=0; j<N; j+=2)
      {
        fprintf(fp,"(%d,%d),\n",matrix_destinations[k+N][0]-N,matrix_destinations[k+N][0]);
        fprintf(fp,"(%d,%d),\n",matrix_destinations[k+N][1]-N,matrix_destinations[k+N][1]);
        k += 2;
      }
  }
      fclose(fp);

}



void print_matrix(int* m, int n, int N)
{
    int i,j;

    for (i=0;i<N;i++)
        {for(j=0;j<n;j++)
            {
                printf("%d\t",m[N*j+i]);
            }
        printf("\n");
        }
    printf("\n");
}
