#ifndef _RESULTS_H
#define _RESULTS_H

#include"params.h"

/// measure the ber and the fer
///\param K is the block size
///\param *dec is the hard decision couple
///\param *frame is the original uncoded frame
///\param **bit_err is the array where bit error counters are stored
///\param **frame_err is the array where frame error counters are stored
///\param itc is the current iteration
///\param curr_point is the current SNR point
///\return the number of error counted for current frame
int BER_meter(int K, int *dec, int *frame, int **bit_err, int **frame_err, int itc, int curr_point);

/// stores the simulation results (ber/fer)
///\param *fp is the results file pointer
///\param min_iter is the minimun number of iterations
///\param max_iter is the maximum number of iterations
///\param max_point_comp is the maximum SNR point computed so far
///\param ebn0_min is the minimum SNR value simulated
///\param ebn0_step is the SNR step between consecutive SNR points
///\param **bit_err is the array where bit error counters are stored
///\param **frame_err is the array where frame error counters are stored
///\param *tot_frames is the array where the total number of frames simulated per each SNR point is stored
///\param K is the frame size
///\param *avg_iter is the average number of iterations
void store_results(FILE *fp, int min_iter, int max_iter, int max_point_comp, double ebn0_min, double ebn0_step, int **bit_err, int **frame_err, int *tot_frames, int K, int *avg_iter, int *itReach);

/// save simulation status for future resuming
///\param lfsr_seed is the seed of the lfsr used to randomly generate uncoded bits
///\param unif_seed is the seed of the uniform generator used to produce gaussian noise samples
///\param curr_frame is the current frame number (for resuming simulation)
///\param max_point_comp is the maximum simulation point computed so far
///\param min_point minimum current starting point
///\param *tot_frames is the total number of simulated frame array
///\param **frame_err is the frame error counter array
///\param **bit_err is the bit error counter array
///\param *avg_iter is the average iteration counter array
void save_simulation(params_t sim_par, unsigned int lfsr_seed, unsigned int unif_seed, int curr_frame, int max_point_comp, int min_point, int *tot_frames, int **frame_err, int **bit_err, int *avg_iter);

///store simulation result in details_ sent and decoded frame at each Eb/N0 and the number of iterations needed for that frame
///\param *fp is the results file pointer
///\param curr_iter is the current iteration
///\param K is the frame size
///\param *decodedK is the hard-decoded frame
///\param *uncodedK is the sent frame
///\param curr_point is the current SNR point
///\param max_point is the maximum number of SNR point
///\param ebn0_min is the minimum SNR value simulated
///\param ebn0_step is the SNR step between consecutive SNR points
///\param curr_frame is the current frame number
///\param *tmp copy uncodedK to check if a desynchronization occours
void save_details(FILE *fp,int curr_iter,int K,int *decodedK,int *uncodedK, int curr_point, int max_point, double ebn0_min, double ebn0_step,int curr_frame, int *tmp, int min_point);

#ifdef FLOAT
void save_hw_check_out(FILE *fp,double *left, double *right,int nN,int N);
void save_hw_check_out_reduced(FILE *fp,double *left, double *right,int nN,int N);

void save_hw_check_llr(FILE *fp, double *llr_in,double *right,int N);

void save_hw_left_or_right(FILE *fp,double *left, int n,int N);
#else
void save_hw_check_out(FILE *fp,int *left, int *right,int nN,int N);
void save_hw_check_out_reduced(FILE *fp,int *left, int *right,int nN,int N);

void save_hw_check_llr(FILE *fp, int *llr_in,int *right,int N);

void save_hw_left_or_right(FILE *fp,int *left, int n,int N);
#endif // FLOAT
void save_hw_check_matrix(FILE *fp, int **matrix_destinations,int N, int n);

void print_matrix(int* m, int n, int N);
#endif
