#include "rv.h"
#include <math.h>
#include <stdlib.h>
#define M_PI 3.14159265358979323846
/*! \brief Generates Random Variables.
 *  \author Erdal Arıkan
 */

/**
Generates uniform random variable.
\return Uniform random variable.
*/
double RVuniform(void)
{
	return ((rand() + 0.5) / (1.0 + RAND_MAX));
}

/**
Generates exponential random variable.
\return Exponential random variable.
*/
double RVexp(void)
{
	return (-log(RVuniform()));
}

double RVnormal(void)
{
	static int s = 0;
	static double r, t;

	if ((s = 1 - s)) {
		r = sqrt(2 * RVexp());
		t = 2 * M_PI * RVuniform();
		return (r * cos(t));
	} else
		return (r * sin(t));
}

void RVprob(double *p, int n)
{
        double s = 1, t;

        if (n <= 0)
                return;

        while (--n > 0) {
                t = exp(-RVexp()/n);
                p[n] = (1-t)*s;
                s *= t;
        }
        p[0] = s;
}

