library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use IEEE.math_real.all;

entity FBER is
  
generic(N 		: integer :=8);
port( clk,hresetn,start,load_new : in std_logic;
		L,X	: in  std_logic_vector(N-1 downto 0);
		DONE_FBER, rest, ENABLE_W : out std_logic);
		
end entity FBER;

architecture rtl of FBER is

component FBER_EU is 
  generic (N 		: integer := 8);

port( clk, hresetn, sresetn, ENABLE_W, sel, ENABLE_S, ENABLE_X, data_valid, ENABLE_O: in std_logic;
      L,X: in std_logic_vector(N-1 downto 0);
		DONE_FBER: out std_logic;
		output_COUNT: out unsigned(2 downto 0));
end component;

component FBER_CU is    
   generic(N : integer := 8);
	
   port( clk : in std_logic;
		   start, load_new : in std_logic;
		   hresetn : in std_logic;
		   sresetn: out std_logic;
		   ENABLE_W : out std_logic;
		   ENABLE_X: out std_logic;
		   sel: out std_logic;
		   ENABLE_S : out std_logic;
		   DATA_VALID : out std_logic;
		   ENABLE_O: out std_logic;
		   n_load: in unsigned(1 downto 0);
		   ENABLE_LOAD: out std_logic;
		   output_COUNT: in unsigned(2 downto 0);
		   rest: out std_logic);
end component;

component counter_up2 is 
    port(	clk,resetn,ENABLE_LOAD:	in std_logic;
	       q	:	out unsigned(1 downto 0)); 
end component;

signal s_output_COUNT2: unsigned(2 downto 0);
signal s_n_load: unsigned(1 downto 0);
signal s_L, s_X	 : std_logic_vector(N-1 downto 0);
signal s_ENABLE_S, s_ENABLE_O, s_ENABLE_OP, s2_sel, s_ENABLE_X, s_data_valid: std_logic;
signal s_DONE_FBER, s_ENABLE_W, sresetn: std_logic;
signal s_load_new, s_rest, s_ENABLE_LOAD: std_logic;

begin
    
DONE_FBER<=s_DONE_FBER;
rest<=s_rest;

EU : FBER_EU generic map(N)
		port map(
		   clk => clk,
		   hresetn => hresetn,
		   sresetn => sresetn,
		   ENABLE_W => s_ENABLE_W,
		   sel => s2_sel,
		   ENABLE_S => s_ENABLE_S,
		   ENABLE_X => s_ENABLE_X,
		   data_valid => s_data_valid,
		   ENABLE_O => s_ENABLE_O,
		   L => L,
		   X => X,
		   output_COUNT => s_output_COUNT2,
		   DONE_FBER => s_DONE_FBER);

CU : FBER_CU generic map(N)
		port map(
		   clk => clk,
		   start => start,
		   load_new => load_new,
		   hresetn => hresetn,
		   sresetn => sresetn,
		   ENABLE_W => s_ENABLE_W,
		   sel => s2_sel,
		   ENABLE_X => s_ENABLE_X,
		   ENABLE_S => s_ENABLE_S, 
		   data_valid => s_data_valid,
		   ENABLE_O => s_ENABLE_O,
		   n_load => s_n_load,
		   ENABLE_LOAD => s_ENABLE_LOAD,
		   output_COUNT => s_output_COUNT2,
		   rest => s_rest);

COUNT2: counter_up2 port map(
         clk => clk,
         resetn => sresetn,
         ENABLE_LOAD => s_ENABLE_LOAD,
         q => s_n_load);

end architecture rtl;
