library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use ieee.numeric_std.all;

entity FBER_CU is 
generic(
	N : integer := 8);
	
port( clk : in std_logic;
		start, load_new : in std_logic;
		hresetn : in std_logic;
		sresetn: out std_logic;
		ENABLE_W : out std_logic;
		sel: out std_logic;
		ENABLE_S : out std_logic;
		ENABLE_X: out std_logic;
		DATA_VALID : out std_logic;
		ENABLE_O : out std_logic;
		ENABLE_LOAD: out std_logic;
		n_load: in unsigned(1 downto 0);
		output_COUNT: in unsigned(2 downto 0);
		rest: out std_logic);
end entity FBER_CU;

architecture behavior of FBER_CU is

type state_type is (IDLE1,INIT1,INIT2,IDLE2,ELAB1,ELAB2,ELAB3,CONTROL,IDLE3,LOAD1,LOAD2); --defintion of state machine type
signal actual_s,next_s: state_type;
signal s1_sel: std_logic:='0';

begin
sel<=s1_sel;
  process (clk,hresetn) is
  begin

	if hresetn='0' then 
		--default
		sresetn <= '0';
	elsif rising_edge(clk) then
	   sresetn <='1';
		actual_s <= next_s;
		ENABLE_W<='0';
		ENABLE_X<='0';
		ENABLE_S<='0';
		DATA_VALID<='0';
		ENABLE_O<='0';
		ENABLE_LOAD<='0';
		
        case next_s is 
            when IDLE1 =>
				if(start= '1') then
                    next_s <= INIT1;
				else
					next_s <= IDLE1;
					sresetn <= '0';
					rest <= '1';
					s1_sel <= '0';
            end if;
		
            when INIT1 =>
                ENABLE_W <= '1';
                rest <= '0';
                ENABLE_LOAD<='1';
                next_s <= INIT2;
            
            when INIT2 =>
                rest<= '0';
                ENABLE_W <= '0';
                ENABLE_S <= '1';
                s1_sel <= not s1_sel;
                if (n_load= "10") then
                next_s <= ELAB1;
                else
                next_s <= IDLE2;
            end if;
            
            when IDLE2 =>
                rest <= '1';
                ENABLE_S <= '0';
                if (load_new= '1') then
                next_s <= INIT1;
                else
                next_s <= IDLE2;
               end if;
                
           when ELAB1 =>
                  next_s <= ELAB2;
                  ENABLE_S <= '0';
                  ENABLE_X <= '1';
                                  
		     when ELAB2 =>
                  next_s <= ELAB3;
                  DATA_VALID <= '1';
                  ENABLE_X <= '0';
                  
	        when ELAB3 =>
                  DATA_VALID <= '0';
                  ENABLE_O <= '1';
                  next_s <= CONTROL;
           
           when CONTROL =>
                  ENABLE_O <= '0';
                  if output_COUNT= "100" then
                               next_s <= IDLE1;
                          else
                               next_s <= IDLE3;
                      end if;
	        
	        when IDLE3 =>
               rest <= '1';
               if (load_new= '1') then
               next_s <= LOAD1;
               else
               next_s <= IDLE3;
           end if;
	        
            when LOAD1 =>
                ENABLE_W <= '1';
                rest <= '0';
                next_s <= LOAD2;
            
            when LOAD2 =>
                ENABLE_S <= '1';
                ENABLE_W <= '0';
                s1_sel <= not s1_sel;
                next_s <= ELAB1;
                      
            when others =>
				sresetn <= '0';
				next_s <= IDLE1;
        end case;
    end if;
end process;    

end architecture behavior;
