library ieee ;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity FBER_EU is

generic(N: natural :=8);

port( clk, hresetn, sresetn, ENABLE_W, sel, ENABLE_S, ENABLE_X, data_valid, ENABLE_O: in std_logic;
      L,X: in std_logic_vector(N-1 downto 0);
		DONE_FBER: out std_logic;
		output_COUNT: out unsigned(2 downto 0));

end entity FBER_EU;

architecture rtl of FBER_EU is	

component input_FBER is

port(	clk,resetn,ENABLE_W: in std_logic;
	   L1,X1: in std_logic_vector(N-1 downto 0);
	   D1: out std_logic_vector(N-1 downto 0));
end component input_FBER;

component xnorN is
    
PORT(   clk, resetn, sel, ENABLE_X, ENABLE_S : IN STD_LOGIC;
        D2   : IN STD_LOGIC_VECTOR(N-1 DOWNTO 0);
        S2   : OUT STD_LOGIC_VECTOR(N-1 DOWNTO 0));
	
end component xnorN;

component output_FBER is

port( clk, resetn, data_valid: in std_logic;
      D3: in std_logic_vector(N-1 downto 0);
      S3: out std_logic);

end component output_FBER;

component control_FBER is

port(	clk,resetn,ENABLE_O,D4:	in std_logic;
	   DONE_FBER: out std_logic;
	   output_COUNT: out unsigned(2 downto 0));
end component control_FBER;

signal resetn,A_1,output_DONE_FBER: std_logic;
signal w_D,w_S: std_logic_vector(N-1 downto 0);
signal s_output_COUNT:unsigned(2 downto 0);

begin

resetn <= hresetn or sresetn;
output_COUNT <= s_output_COUNT;
DONE_FBER <= output_DONE_FBER;

FBER_1: input_FBER port map(
            clk => clk,
            resetn => resetn,
            ENABLE_W => ENABLE_W,
            L1 => L,
            X1 => X,
            D1 => w_D);

FBER_2: xnorN port map(
            clk => clk,
            resetn => resetn,
            sel => sel,
            ENABLE_X => ENABLE_X,
            ENABLE_S => ENABLE_S,
            D2 => w_D,
            S2 => w_S);

FBER_3: output_FBER port map(
            clk => clk,
            resetn => resetn,
            data_valid => data_valid,
            D3 => w_S,
            S3 => A_1);

FBER_4: control_FBER port map(
            clk => clk,
            resetn => resetn,
            ENABLE_O => ENABLE_O,
            D4 => A_1,
            DONE_FBER => output_DONE_FBER,
            output_COUNT => s_output_COUNT);

end architecture rtl;
