#include"myservs.h"
#include"encoder.h"

/// Create G (generator) matrix
///\param **g is the matrix pointer
///\param n is log2(N)
void create_g(int **G, int n)
{
  int l;
  int N;
  int i,j;

  /// init G with G2
  G[0][0] = 1;
  G[0][1] = 0;
  G[1][0] = 1;
  G[1][1] = 1;

  for (l=1; l<n; l++)
  {
    N = 1 << l;
    /// put zeros in the second quad
    for (i=0; i<N; i++)
      for (j=N; j<2*N; j++)
	G[i][j] = 0;
    /// copy the first quad in the third and fourth
    for (i=N; i<2*N; i++)
      for (j=0; j<N; j++)
	G[i][j] = G[i-N][j];
    for (i=N; i<2*N; i++)
      for (j=N; j<2*N; j++)
	G[i][j] = G[i-N][j-N];
  }
}

/// Transpose G to speed up encoding
///\param **G is the matrix pointer
///\param **Gt is the transposed matrix pointer
///\param N is the matrix size
void transpose_g(int **G, int **Gt, int N)
{
  int i,j;

  for (i=0; i<N; i++)
    for (j=0; j<N; j++)
      Gt[i][j] = G[j][i];
}

/// encoding function
///\param *uncoded is the uncoded array
///\param *coded is the coded array
///\param **Gt is the transposed generator matrix
///\param N is the array length
void encode(int *uncoded, int *coded, int **Gt, int N)
{
  int i,j;
  /// exploit triangular lower matrix structure

  for (i=0; i<N; i++)
  {
    coded[i] = uncoded[N-1];
    for (j=N-2; j>=i; j--)
      coded[i] ^= (Gt[i][j] & uncoded[j]);
  }
}
